import asyncio
from unittest.mock import AsyncMock, MagicMock, Mock

import pytest
from bluesky.protocols import Reading

from ophyd_async.core import (
    CALCULATE_TIMEOUT,
    AsyncStatus,
    FlyMotorInfo,
    init_devices,
    observe_value,
    soft_signal_rw,
)
from ophyd_async.epics import motor
from ophyd_async.testing import (
    StatusWatcher,
    callback_on_mock_put,
    get_mock_put,
    mock_puts_blocked,
    set_mock_put_proceeds,
    set_mock_value,
    wait_for_pending_wakeups,
)


@pytest.fixture
async def sim_motor():
    async with init_devices(mock=True):
        sim_motor = motor.Motor("BLxxI-MO-TABLE-01:X", name="sim_motor")

    set_mock_value(sim_motor.motor_egu, "mm")
    set_mock_value(sim_motor.precision, 3)
    set_mock_value(sim_motor.velocity, 1)
    set_mock_value(sim_motor.low_limit_travel, -10.01)
    set_mock_value(sim_motor.high_limit_travel, 20.01)
    yield sim_motor


@pytest.mark.xfail(reason="Flaky test")
async def test_motor_moving_well(sim_motor: motor.Motor) -> None:
    set_mock_put_proceeds(sim_motor.user_setpoint, False)
    s = sim_motor.set(0.55)
    watcher = StatusWatcher(s)
    await watcher.wait_for_call(
        name="sim_motor",
        current=0.0,
        initial=0.0,
        target=0.55,
        unit="mm",
        precision=3,
        time_elapsed=pytest.approx(0.0, abs=0.1),
    )
    assert 0.55 == await sim_motor.user_setpoint.get_value()
    assert not s.done
    await asyncio.sleep(0.1)
    set_mock_value(sim_motor.user_readback, 0.1)
    await watcher.wait_for_call(
        name="sim_motor",
        current=0.1,
        initial=0.0,
        target=0.55,
        unit="mm",
        precision=3,
        time_elapsed=pytest.approx(0.1, abs=0.1),
    )
    set_mock_put_proceeds(sim_motor.user_setpoint, True)
    await wait_for_pending_wakeups()
    assert s.done


async def test_motor_move_timeout(sim_motor: motor.Motor):
    class MyError(Exception):
        pass

    def do_timeout(value, wait):
        # Raise custom exception to be clear it bubbles up
        raise MyError()

    callback_on_mock_put(sim_motor.user_setpoint, do_timeout)
    s = sim_motor.set(0.3)
    watcher = Mock()
    s.watch(watcher)
    with pytest.raises(MyError):
        await s
    watcher.assert_called_once_with(
        name="sim_motor",
        current=0.0,
        initial=0.0,
        target=0.3,
        unit="mm",
        precision=3,
        time_elapsed=pytest.approx(0.0, abs=0.2),
    )


async def test_motor_moving_stopped(sim_motor: motor.Motor):
    set_mock_value(sim_motor.motor_done_move, False)
    set_mock_put_proceeds(sim_motor.user_setpoint, False)
    s = sim_motor.set(1.5)
    s.add_callback(Mock())
    await asyncio.sleep(0.2)
    assert not s.done
    await sim_motor.stop()

    # Note: needs to explicitly be called with 1, not just processed.
    # See https://epics.anl.gov/bcda/synApps/motor/motorRecord.html#Fields_command
    get_mock_put(sim_motor.motor_stop).assert_called_once_with(1, wait=False)

    set_mock_put_proceeds(sim_motor.user_setpoint, True)
    await wait_for_pending_wakeups()
    assert s.done
    assert s.success is False


async def test_read_motor(sim_motor: motor.Motor):
    await sim_motor.stage()
    assert (await sim_motor.read())["sim_motor"]["value"] == 0.0
    assert (await sim_motor.read_configuration())["sim_motor-velocity"]["value"] == 1
    assert (await sim_motor.describe_configuration())["sim_motor-motor_egu"][
        "shape"
    ] == []
    set_mock_value(sim_motor.user_readback, 0.5)
    assert (await sim_motor.read())["sim_motor"]["value"] == 0.5
    await sim_motor.unstage()
    # Check we can still read and describe when not staged
    set_mock_value(sim_motor.user_readback, 0.1)
    assert (await sim_motor.read())["sim_motor"]["value"] == 0.1
    assert await sim_motor.describe()


async def test_set_velocity(sim_motor: motor.Motor) -> None:
    v = sim_motor.velocity
    q: asyncio.Queue[dict[str, Reading]] = asyncio.Queue()
    v.subscribe_reading(q.put_nowait)
    assert (await q.get())["sim_motor-velocity"]["value"] == 1.0
    await v.set(-2.0)
    assert (await q.get())["sim_motor-velocity"]["value"] == -2.0
    v.clear_sub(q.put_nowait)
    await v.set(3.0)
    assert (await v.read())["sim_motor-velocity"]["value"] == 3.0
    assert q.empty()


async def test_set_with_zero_velocity(sim_motor: motor.Motor) -> None:
    await sim_motor.velocity.set(0)
    with pytest.raises(ValueError, match="Mover has zero velocity"):
        await sim_motor.set(3.14)


@pytest.mark.parametrize(
    "position, upper_limit, lower_limit",
    [
        (-10, 9.99, -9.99),  # Goes below lower_limit
        (10, 9.99, -9.99),  # Goes above upper_limit
    ],
)
async def test_move_outside_motor_limits_causes_error(
    sim_motor: motor.Motor,
    position,
    upper_limit,
    lower_limit,
):
    set_mock_value(sim_motor.velocity, 10)
    set_mock_value(sim_motor.low_limit_travel, lower_limit)
    set_mock_value(sim_motor.high_limit_travel, upper_limit)
    with pytest.raises(motor.MotorLimitsError):
        await sim_motor.set(position)


async def test_given_limits_of_0_0_then_move_causes_no_error(
    sim_motor: motor.Motor,
):
    set_mock_value(sim_motor.velocity, 10)
    set_mock_value(sim_motor.low_limit_travel, 0)
    set_mock_value(sim_motor.high_limit_travel, 0)
    await sim_motor.set(100)
    assert (await sim_motor.user_setpoint.get_value()) == 100


@pytest.mark.parametrize(
    "setpoint, velocity, timeout",
    [
        (1, 1, CALCULATE_TIMEOUT),
        (-1, -1, CALCULATE_TIMEOUT),
        (1, -1, CALCULATE_TIMEOUT),
        (-1, 1, CALCULATE_TIMEOUT),
        (1, 1, 1),
        (-1, -1, 1),
        (1, -1, 1),
        (-1, 1, 1),
    ],
)
async def test_set(sim_motor: motor.Motor, setpoint, velocity, timeout) -> None:
    await sim_motor.velocity.set(velocity)
    await sim_motor.set(setpoint, timeout=timeout)
    assert (await sim_motor.locate()).get("setpoint") == setpoint


async def test_prepare_velocity_limit_error(sim_motor: motor.Motor):
    set_mock_value(sim_motor.max_velocity, 10)
    with pytest.raises(motor.MotorLimitsError):
        fly_info = FlyMotorInfo(start_position=-10, end_position=0, time_for_move=0.9)
        await sim_motor.prepare(fly_info)


async def test_valid_prepare_velocity(sim_motor: motor.Motor):
    set_mock_value(sim_motor.low_limit_travel, -10.01)
    set_mock_value(sim_motor.high_limit_travel, 20.01)
    set_mock_value(sim_motor.max_velocity, 10)
    fly_info = FlyMotorInfo(start_position=-10, end_position=0, time_for_move=1)
    await sim_motor.prepare(fly_info)
    assert (
        await sim_motor.velocity.get_value() == await sim_motor.max_velocity.get_value()
    )


@pytest.mark.parametrize(
    "acceleration_time, velocity, start_position, end_position, upper_limit,\
    lower_limit",
    [
        (1, 10, 0, 10, 30, -4.999),  # Goes below lower_limit, +ve direction
        (1, 10, 0, 10, 14.99, -10),  # Goes above upper_limit, +ve direction
        (1, 10, 10, 0, -30, -9.999),  # Goes below lower_limit, -ve direction
        (1, 10, 10, 0, 14.99, -10),  # Goes above upper_limit, -ve direction
    ],
)
async def test_prepare_motor_limits_error(
    sim_motor: motor.Motor,
    acceleration_time,
    velocity,
    start_position,
    end_position,
    upper_limit,
    lower_limit,
):
    set_mock_value(sim_motor.acceleration_time, acceleration_time)
    set_mock_value(sim_motor.low_limit_travel, lower_limit)
    set_mock_value(sim_motor.high_limit_travel, upper_limit)
    time_for_move = abs(end_position - start_position) / velocity
    fly_info = FlyMotorInfo(
        start_position=start_position,
        end_position=end_position,
        time_for_move=time_for_move,
    )
    with pytest.raises(motor.MotorLimitsError):
        await sim_motor.prepare(fly_info)


async def test_prepare_valid_limits(sim_motor: motor.Motor):
    set_mock_value(sim_motor.acceleration_time, 1)
    set_mock_value(sim_motor.low_limit_travel, -10.01)
    set_mock_value(sim_motor.high_limit_travel, 20.01)
    set_mock_value(sim_motor.max_velocity, 10)
    fly_info = motor.FlyMotorInfo(
        start_position=0,
        end_position=10,
        time_for_move=1,
    )
    await sim_motor.prepare(fly_info)
    assert await sim_motor.user_setpoint.get_value() == -5

    assert (
        fly_info.ramp_down_end_pos(await sim_motor.acceleration_time.get_value()) == 15
    )
    fly_info = motor.FlyMotorInfo(
        start_position=12,
        end_position=2,
        time_for_move=1,
    )
    await sim_motor.prepare(fly_info)
    assert await sim_motor.user_setpoint.get_value() == 17
    assert (
        fly_info.ramp_down_end_pos(await sim_motor.acceleration_time.get_value()) == -3
    )


@pytest.mark.parametrize(
    "expected_velocity, target_position",
    [
        (10, -10),
        (8, 8),
    ],
)
async def test_prepare(
    sim_motor: motor.Motor, target_position: float, expected_velocity: float
):
    set_mock_value(sim_motor.acceleration_time, 1)
    set_mock_value(sim_motor.low_limit_travel, -15)
    set_mock_value(sim_motor.high_limit_travel, 20)
    set_mock_value(sim_motor.max_velocity, 10)
    fake_set_signal = soft_signal_rw(float)
    await fake_set_signal.connect()

    async def wait_for_set(_):
        async for value in observe_value(fake_set_signal, timeout=1):
            if value == target_position:
                break

    sim_motor.set = AsyncMock(side_effect=wait_for_set)

    async def do_set(status: AsyncStatus):
        assert not status.done
        await fake_set_signal.set(target_position)

    async def wait_for_status(status: AsyncStatus):
        await status

    status = sim_motor.prepare(
        motor.FlyMotorInfo(
            start_position=0,
            end_position=target_position,
            time_for_move=1,
        )
    )
    # Test that prepare is not marked as complete until correct position is reached
    await asyncio.gather(do_set(status), wait_for_status(status))
    assert await sim_motor.velocity.get_value() == expected_velocity
    assert status.done


async def test_kickoff(sim_motor: motor.Motor):
    sim_motor.set = MagicMock()
    with pytest.raises(
        RuntimeError, match="Motor must be prepared before attempting to kickoff"
    ):
        await sim_motor.kickoff()
    # TODO: why was this called _twice_?
    # with pytest.raises(RuntimeError):
    #     await sim_motor.kickoff()
    set_mock_value(sim_motor.acceleration_time, 1)
    sim_motor._fly_info = motor.FlyMotorInfo(
        start_position=12,
        end_position=2,
        time_for_move=1,
    )
    await sim_motor.kickoff()
    sim_motor.set.assert_called_once_with(-3.0, timeout=CALCULATE_TIMEOUT)


async def test_complete(sim_motor: motor.Motor) -> None:
    with pytest.raises(RuntimeError, match="kickoff not called"):
        sim_motor.complete()
    sim_motor._fly_status = sim_motor.set(20)
    assert not sim_motor._fly_status.done
    await sim_motor.complete()
    assert sim_motor._fly_status.done


async def test_locatable(sim_motor: motor.Motor) -> None:
    callback_on_mock_put(
        sim_motor.user_setpoint,
        lambda x, *_, **__: set_mock_value(sim_motor.user_readback, x),
    )
    assert (await sim_motor.locate())["readback"] == 0
    with mock_puts_blocked(sim_motor.user_setpoint):
        move_status = sim_motor.set(10)
        assert (await sim_motor.locate())["readback"] == 0
    await move_status
    assert (await sim_motor.locate())["readback"] == 10
    assert (await sim_motor.locate())["setpoint"] == 10


def test_core_notconnected_emits_deprecation_warning():
    with pytest.deprecated_call():
        from ophyd_async.epics.motor import MotorLimitsException  # noqa: F401



from unittest.mock import MagicMock

import pytest
from bluesky.protocols import HasHints

from ophyd_async.core import (
    AsyncConfigurable,
    AsyncReadable,
    AsyncStageable,
    ConfigSignal,
    Device,
    DeviceVector,
    HintedSignal,
    MockSignalBackend,
    SignalR,
    StandardReadable,
    soft_signal_r_and_setter,
    soft_signal_rw,
)
from ophyd_async.core import StandardReadableFormat as Format


@pytest.mark.parametrize("wrapper", [HintedSignal, HintedSignal.uncached, ConfigSignal])
def test_standard_readable_wrappers_raise_deprecation_warning(wrapper):
    sr = StandardReadable()
    with pytest.deprecated_call():
        sr.add_readables([soft_signal_rw(int)], wrapper)


def test_standard_readable_hints():
    sr = StandardReadable()

    assert sr.hints == {}

    hint1 = MagicMock(spec=HasHints)
    hint1.hints = {"fields": ["abc"], "dimensions": [(["f1", "f2"], "s1")]}

    hint2 = MagicMock(spec=HasHints)
    hint2.hints = {"fields": ["def", "ghi"]}

    hint3 = MagicMock(spec=HasHints)
    hint3.hints = {"fields": ["jkl"], "gridding": "rectilinear_nonsequential"}

    sr.add_readables([hint1, hint2, hint3])

    assert sr.hints == {
        "fields": ["abc", "def", "ghi", "jkl"],
        "dimensions": [(["f1", "f2"], "s1")],
        "gridding": "rectilinear_nonsequential",
    }


def test_standard_readable_hints_raises_when_overriding_string_literal():
    sr = StandardReadable()

    hint1 = MagicMock(spec=HasHints)
    hint1.hints = {"gridding": "rectilinear_nonsequential"}

    hint2 = MagicMock(spec=HasHints)
    hint2.hints = {"gridding": "a different string"}

    sr._has_hints = (
        hint1,
        hint2,
    )

    with pytest.raises(RuntimeError, match=r"Hints key .* value may not be overridden"):
        sr.hints  # noqa: B018


def test_standard_readable_hints_raises_when_overriding_sequence():
    sr = StandardReadable()

    hint1 = MagicMock(spec=HasHints)
    hint1.hints = {"fields": ["field1", "field2"]}

    hint2 = MagicMock(spec=HasHints)
    hint2.hints = {"fields": ["field2"]}

    sr._has_hints = (
        hint1,
        hint2,
    )

    with pytest.raises(RuntimeError, match=r"Hint fields .* overrides existing hint"):
        sr.hints  # noqa: B018


@pytest.mark.parametrize("invalid_type", [1, 1.0, {"abc": "def"}, {1, 2, 3}])
def test_standard_readable_hints_invalid_types(invalid_type):
    sr = StandardReadable()

    hint1 = MagicMock(spec=HasHints)
    hint1.hints = {"test": invalid_type}

    sr._has_hints = (hint1,)

    with pytest.raises(TypeError, match=r"Unknown type for value .* for key .*"):
        sr.hints  # noqa: B018


def test_standard_readable_add_children_context_manager():
    sr = StandardReadable()
    mock = MagicMock()
    sr.add_readables = mock
    with sr.add_children_as_readables():
        sr.a = MagicMock(spec=SignalR)
        sr.b = MagicMock(spec=SignalR)
        sr.c = MagicMock(spec=SignalR)

    # Can't use assert_called_once_with() as the order of items returned from
    # internal dict comprehension is not guaranteed
    mock.assert_called_once()
    assert set(mock.call_args.args[0]) == {sr.a, sr.b, sr.c}


def test_standard_readable_add_children_cm_device_vector():
    sr = StandardReadable()
    mock = MagicMock()
    sr.add_readables = mock

    # Create a mock for the DeviceVector.children() call
    mock_d1 = MagicMock(spec=SignalR)
    mock_d2 = MagicMock(spec=SignalR)
    mock_d3 = MagicMock(spec=SignalR)
    vector = DeviceVector({1: mock_d1, 2: mock_d2, 3: mock_d3})
    with sr.add_children_as_readables():
        sr.a = vector

    # Can't use assert_called_once_with() as the order of items returned from
    # internal dict comprehension is not guaranteed
    mock.assert_called_once()
    assert set(mock.call_args.args[0]) == {mock_d1, mock_d2, mock_d3}


def test_standard_readable_add_children_cm_filters_non_devices():
    sr = StandardReadable()
    mock = MagicMock()
    sr.add_readables = mock

    with sr.add_children_as_readables():
        sr.a = MagicMock(spec=SignalR)
        sr.b = MagicMock(spec=Device)
        sr.c = 1.0
        sr.d = "abc"
        sr.e = MagicMock(spec=MockSignalBackend)

    # Can't use assert_called_once_with() as the order of items returned from
    # internal dict comprehension is not guaranteed
    mock.assert_called_once()
    assert set(mock.call_args.args[0]) == {sr.a, sr.b}


def assert_sr_has_attrs(sr: StandardReadable, expected_attrs: dict[str, tuple]):
    attrs_to_check = (
        "_describe_config_funcs",
        "_read_config_funcs",
        "_describe_funcs",
        "_read_funcs",
        "_stageables",
        "_has_hints",
    )
    actual = {attr: getattr(sr, attr) for attr in attrs_to_check}
    expected = {attr: expected_attrs.get(attr, ()) for attr in attrs_to_check}
    assert actual == expected


signal_r = MagicMock(spec=SignalR)
async_readable = MagicMock(spec=AsyncReadable)
async_configurable = MagicMock(spec=AsyncConfigurable)
async_stageable = MagicMock(spec=AsyncStageable)
has_hints = MagicMock(spec=HasHints)


@pytest.mark.parametrize(
    "readable, expected_attrs",
    [
        (
            signal_r,
            {
                "_read_funcs": (signal_r.read,),
                "_describe_funcs": (signal_r.describe,),
                "_stageables": (signal_r,),
            },
        ),
        (
            async_readable,
            {
                "_read_funcs": (async_readable.read,),
                "_describe_funcs": (async_readable.describe,),
            },
        ),
        (
            async_configurable,
            {
                "_read_config_funcs": (async_configurable.read_configuration,),
                "_describe_config_funcs": (async_configurable.describe_configuration,),
            },
        ),
        (async_stageable, {"_stageables": (async_stageable,)}),
        (has_hints, {"_has_hints": (has_hints,)}),
    ],
)
def test_standard_readable_add_readables_adds_to_expected_attrs(
    readable, expected_attrs: dict[str, tuple]
):
    sr = StandardReadable()
    sr.add_readables([readable])
    assert_sr_has_attrs(sr, expected_attrs)


@pytest.mark.parametrize(
    "format",
    [
        Format.CONFIG_SIGNAL,
        Format.HINTED_SIGNAL,
        Format.UNCACHED_SIGNAL,
        Format.HINTED_UNCACHED_SIGNAL,
    ],
)
def test_standard_readable_add_readables_raises_signalr_typeerror(format) -> None:
    # Mock a Device instance that is not a SignalR
    mock_device = MagicMock(spec=Device)
    sr = StandardReadable()

    # Ensure it raises TypeError
    with pytest.raises(TypeError, match=f"{mock_device} is not a SignalR"):
        sr.add_readables([mock_device], format=format)


def test_standard_readable_config_signal():
    signal_r = MagicMock(spec=SignalR)
    sr = StandardReadable()
    sr.add_readables([signal_r], Format.CONFIG_SIGNAL)
    assert sr._describe_config_funcs == (signal_r.describe,)
    assert sr._read_config_funcs == (signal_r.read,)


def test_standard_readable_hinted_signal():
    signal_r = MagicMock(spec=SignalR)
    sr = StandardReadable()
    sr.add_readables([signal_r], Format.HINTED_SIGNAL)
    assert sr._describe_funcs == (signal_r.describe,)
    assert sr._read_funcs == (signal_r.read,)
    assert sr._stageables == (signal_r,)
    assert sr._has_hints[0].device == signal_r


def test_standard_readable_uncached_signal():
    signal_r = MagicMock(spec=SignalR)
    sr = StandardReadable()
    sr.add_readables([signal_r], Format.UNCACHED_SIGNAL)
    assert sr._describe_funcs == (signal_r.describe,)
    assert sr._read_funcs[0].signal == signal_r


def test_standard_readable_hinted_uncached_signal():
    signal_r = MagicMock(spec=SignalR)
    sr = StandardReadable()
    sr.add_readables([signal_r], Format.HINTED_UNCACHED_SIGNAL)
    assert sr._describe_funcs == (signal_r.describe,)
    assert sr._read_funcs[0].signal == signal_r
    assert sr._has_hints[0].device == signal_r


def test_standard_readable_add_children_multi_nested():
    inner = StandardReadable()
    outer = StandardReadable()
    with inner.add_children_as_readables(Format.HINTED_SIGNAL):
        inner.a, _ = soft_signal_r_and_setter(float, initial_value=5.0)
        inner.b, _ = soft_signal_r_and_setter(float, initial_value=6.0)
    with outer.add_children_as_readables():
        outer.inner = inner
    assert outer


async def test_duplicate_readable_raises_exception():
    class DummyBaseDevice(StandardReadable):
        def __init__(self, name):
            with self.add_children_as_readables():
                self.twin = soft_signal_rw(float)
            super().__init__(name)

    class DummyDerivedDevice(DummyBaseDevice):
        def __init__(self, name):
            with self.add_children_as_readables():
                self.twin = soft_signal_rw(float)
            super().__init__(name)

    with pytest.raises(KeyError):
        DummyDerivedDevice("test_duplicates")



import pytest

from ophyd_async.plan_stubs import setup_ndstats_sum
from ophyd_async.testing import ParentOfEverythingDevice


@pytest.fixture
async def parent_device() -> ParentOfEverythingDevice:
    device = ParentOfEverythingDevice("parent")
    await device.connect(mock=True)
    return device


def test_setup_ndstats_raises_type_error(RE, parent_device: ParentOfEverythingDevice):
    detector = parent_device
    detector_name = "faulty_det"
    detector.set_name(detector_name)
    with pytest.raises(
        TypeError,
        match=(
            f"Expected {detector_name} to have 'fileio' attribute that is an "
            "NDFileHDFIO"
        ),
    ):
        RE(setup_ndstats_sum(detector))



from typing import Annotated as A

from ophyd_async.core import SignalR, SignalRW, StrictEnum
from ophyd_async.epics import adcore
from ophyd_async.epics.core import (
    PvSuffix,
)


class Andor2TriggerMode(StrictEnum):
    INTERNAL = "Internal"
    EXT_TRIGGER = "External"
    EXT_START = "External Start"
    EXT_EXPOSURE = "External Exposure"
    EXT_FVP = "External FVP"
    SOFTWARE = "Software"


class Andor2DriverIO(adcore.ADBaseIO):
    """Driver for andor model:DU897_BV as deployed on p99.

    This mirrors the interface provided by AdAndor/db/andor.template.
    https://areadetector.github.io/areaDetector/ADAndor/andorDoc.html
    """

    trigger_mode: A[SignalRW[Andor2TriggerMode], PvSuffix.rbv("TriggerMode")]
    andor_accumulate_period: A[SignalR[float], PvSuffix("AndorAccumulatePeriod_RBV")]



from __future__ import annotations

import asyncio
import logging
from collections.abc import Awaitable, Callable, Iterable, Mapping, Sequence
from dataclasses import dataclass
from enum import Enum, EnumMeta, StrEnum
from typing import (
    Any,
    Generic,
    Literal,
    ParamSpec,
    TypeVar,
    get_args,
    get_origin,
)
from unittest.mock import Mock

import numpy as np
from pydantic import BaseModel, ConfigDict

T = TypeVar("T")
V = TypeVar("V")
P = ParamSpec("P")
Callback = Callable[[T], None]
DEFAULT_TIMEOUT = 10.0

logger = logging.getLogger("ophyd_async")


class UppercaseNameEnumMeta(EnumMeta):
    def __new__(cls, *args, **kwargs):
        ret = super().__new__(cls, *args, **kwargs)
        lowercase_names = [x.name for x in ret if not x.name.isupper()]  # type: ignore
        if lowercase_names:
            raise TypeError(f"Names {lowercase_names} should be uppercase")
        return ret


class AnyStringUppercaseNameEnumMeta(UppercaseNameEnumMeta):
    def __call__(cls, value, *args, **kwargs):  # type: ignore
        """Return given value if it is a string and not a member of the enum.

        If the value is not a string or is an enum member, default enum behavior
        is applied. Type checking will complain if provided arbitrary string.

        Returns:
            Union[str, SubsetEnum]: If the value is a string and not a member of the
            enum, the string is returned as is. Otherwise, the corresponding enum
            member is returned.

        Raises:
            ValueError: If the value is not a string and cannot be converted to an enum
            member.

        """
        if isinstance(value, str) and not isinstance(value, cls):
            return value
        return super().__call__(value, *args, **kwargs)


class StrictEnum(StrEnum, metaclass=UppercaseNameEnumMeta):
    """All members should exist in the Backend, and there will be no extras."""


class SubsetEnum(StrEnum, metaclass=AnyStringUppercaseNameEnumMeta):
    """All members should exist in the Backend, but there may be extras."""


class SupersetEnum(StrEnum, metaclass=UppercaseNameEnumMeta):
    """Some members should exist in the Backend, and there should be no extras."""


EnumTypes = StrictEnum | SubsetEnum | SupersetEnum


CALCULATE_TIMEOUT = "CALCULATE_TIMEOUT"
"""Sentinel used to implement ``myfunc(timeout=CalculateTimeout)``

This signifies that the function should calculate a suitable non-zero
timeout itself
"""


CalculatableTimeout = float | None | Literal["CALCULATE_TIMEOUT"]


class NotConnectedError(Exception):
    """Exception to be raised if a `Device.connect` is cancelled.

    :param errors:
        Mapping of device name to Exception or another NotConnectedError.
        Alternatively a string with the signal error text.
    """

    _indent_width = "    "

    def __init__(self, errors: str | Mapping[str, Exception]):
        self._errors = errors

    @property
    def sub_errors(self) -> Mapping[str, Exception]:
        if isinstance(self._errors, dict):
            return self._errors.copy()
        else:
            return {}

    def _format_sub_errors(self, name: str, error: Exception, indent="") -> str:
        if isinstance(error, NotConnectedError):
            error_txt = ":" + error.format_error_string(indent + self._indent_width)
        elif isinstance(error, Exception):
            error_txt = ": " + err_str + "\n" if (err_str := str(error)) else "\n"
        else:
            raise RuntimeError(
                f"Unexpected type `{type(error)}`, expected an Exception"
            )

        string = f"{indent}{name}: {type(error).__name__}" + error_txt
        return string

    def format_error_string(self, indent="") -> str:
        if not isinstance(self._errors, dict) and not isinstance(self._errors, str):
            raise RuntimeError(
                f"Unexpected type `{type(self._errors)}` expected `str` or `dict`"
            )

        if isinstance(self._errors, str):
            return " " + self._errors + "\n"

        string = "\n"
        for name, error in self._errors.items():
            string += self._format_sub_errors(name, error, indent=indent)
        return string

    def __str__(self) -> str:
        return self.format_error_string(indent="")

    @classmethod
    def with_other_exceptions_logged(
        cls, exceptions: Mapping[str, Exception]
    ) -> NotConnectedError:
        for name, exception in exceptions.items():
            if not isinstance(exception, NotConnectedError):
                logger.exception(
                    f"device `{name}` raised unexpected exception "
                    f"{type(exception).__name__}",
                    exc_info=exception,
                )
        return NotConnectedError(exceptions)


@dataclass(frozen=True)
class WatcherUpdate(Generic[T]):
    """A dataclass such that, when expanded, it provides the kwargs for a watcher."""

    current: T
    """The current value, where it currently is."""

    initial: T
    """The initial value, where it was when it started."""

    target: T
    """The target value, where it will be when it finishes."""

    name: str | None = None
    """An optional name for the device, if available."""

    unit: str | None = None
    """Units of the value, if applicable."""

    precision: float | None = None
    """How many decimal places the value should be displayed to."""

    fraction: float | None = None
    """The fraction of the way between initial and target."""

    time_elapsed: float | None = None
    """The time elapsed since the start of the operation."""

    time_remaining: float | None = None
    """The time remaining until the operation completes."""


async def wait_for_connection(**coros: Awaitable[None]):
    """Call many underlying signals, accumulating exceptions and returning them.

    Expected kwargs should be a mapping of names to coroutine tasks to execute.
    """
    exceptions: dict[str, Exception] = {}
    if len(coros) == 1:
        # Single device optimization
        name, coro = coros.popitem()
        try:
            await coro
        except Exception as exc:
            exceptions[name] = exc
    else:
        # Use gather to connect in parallel
        results = await asyncio.gather(*coros.values(), return_exceptions=True)
        for name, result in zip(coros, results, strict=False):
            if isinstance(result, Exception):
                exceptions[name] = result

    if exceptions:
        raise NotConnectedError.with_other_exceptions_logged(exceptions)


def get_dtype(datatype: type) -> np.dtype:
    """Get the runtime dtype from a numpy ndarray type annotation.

    ```python
    >>> from ophyd_async.core import Array1D
    >>> import numpy as np
    >>> get_dtype(Array1D[np.int8])
    dtype('int8')

    ```
    """
    if not get_origin(datatype) == np.ndarray:
        raise TypeError(f"Expected Array1D[dtype], got {datatype}")
    # datatype = numpy.ndarray[typing.Any, numpy.dtype[numpy.float64]]
    # so extract numpy.float64 from it
    return np.dtype(get_args(get_args(datatype)[1])[0])


def get_enum_cls(datatype: type | None) -> type[EnumTypes] | None:
    """Get the enum class from a datatype.

    :raises TypeError: if type is not a [](#StrictEnum) or [](#SubsetEnum)
    or [](#SupersetEnum) subclass
    ```python
    >>> from ophyd_async.core import StrictEnum
    >>> from collections.abc import Sequence
    >>> class MyEnum(StrictEnum):
    ...     A = "A value"
    >>> get_enum_cls(str)
    >>> get_enum_cls(MyEnum)
    <enum 'MyEnum'>
    >>> get_enum_cls(Sequence[MyEnum])
    <enum 'MyEnum'>

    ```
    """
    if get_origin(datatype) is Sequence:
        datatype = get_args(datatype)[0]
    datatype = get_origin_class(datatype)
    if datatype and issubclass(datatype, Enum):
        if not issubclass(datatype, EnumTypes):
            raise TypeError(
                f"{datatype} should inherit from ophyd_async.core.SubsetEnum "
                "or ophyd_async.core.StrictEnum "
                "or ophyd_async.core.SupersetEnum."
            )
        return datatype
    return None


def get_unique(values: dict[str, T], types: str) -> T:
    """If all values are the same, return that value, otherwise raise TypeError.

    ```python
    >>> get_unique({"a": 1, "b": 1}, "integers")
    1
    >>> get_unique({"a": 1, "b": 2}, "integers")
    Traceback (most recent call last):
     ...
    TypeError: Differing integers: a has 1, b has 2

    ```
    """
    set_values = set(values.values())
    if len(set_values) != 1:
        diffs = ", ".join(f"{k} has {v}" for k, v in values.items())
        raise TypeError(f"Differing {types}: {diffs}")
    return set_values.pop()


async def merge_gathered_dicts(
    coros: Iterable[Awaitable[dict[str, T]]],
) -> dict[str, T]:
    """Merge dictionaries produced by a sequence of coroutines.

    Can be used for merging `read()` or `describe()`.

    :example:
    ```python
    combined_read = await merge_gathered_dicts(s.read() for s in signals)
    ```
    """
    ret: dict[str, T] = {}
    for result in await asyncio.gather(*coros):
        ret.update(result)
    return ret


async def gather_dict(coros: Mapping[T, Awaitable[V]]) -> dict[T, V]:
    """Take named coros and return a dict of their name to their return value."""
    values = await asyncio.gather(*coros.values())
    return dict(zip(coros, values, strict=True))


def in_micros(t: float) -> int:
    """Convert between a seconds and microseconds.

    :param t: A time in seconds
    :return: A time in microseconds, rounded up to the nearest whole microsecond
    :raises ValueError: if t < 0
    """
    if t < 0:
        raise ValueError(f"Expected a positive time in seconds, got {t!r}")
    return int(np.ceil(t * 1e6))


def get_origin_class(annotatation: Any) -> type | None:
    origin = get_origin(annotatation) or annotatation
    if isinstance(origin, type):
        return origin
    return None


class Reference(Generic[T]):
    """Hide an object behind a reference.

    Used to opt out of the naming/parent-child relationship of `Device`.

    :example:
    ```python
    class DeviceWithRefToSignal(Device):
        def __init__(self, signal: SignalRW[int]):
            self.signal_ref = Reference(signal)
            super().__init__()

        def set(self, value) -> AsyncStatus:
            return self.signal_ref().set(value + 1)
    ```
    """

    def __init__(self, obj: T):
        self._obj = obj

    def __call__(self) -> T:
        return self._obj


class LazyMock:
    """A lazily created Mock to be used when connecting in mock mode.

    Creating Mocks is reasonably expensive when each Device (and Signal)
    requires its own, and the tree is only used when ``Signal.set()`` is
    called. This class allows a tree of lazily connected Mocks to be
    constructed so that when the leaf is created, so are its parents.
    Any calls to the child are then accessible from the parent mock.

    ```python
    >>> parent = LazyMock()
    >>> child = parent.child("child")
    >>> child_mock = child()
    >>> child_mock()  # doctest: +ELLIPSIS
    <Mock name='mock.child()' id='...'>
    >>> parent_mock = parent()
    >>> parent_mock.mock_calls
    [call.child()]

    ```
    """

    def __init__(self, name: str = "", parent: LazyMock | None = None) -> None:
        self.parent = parent
        self.name = name
        self._mock: Mock | None = None

    def child(self, name: str) -> LazyMock:
        """Return a child of this LazyMock with the given name."""
        return LazyMock(name, self)

    def __call__(self) -> Mock:
        if self._mock is None:
            self._mock = Mock(spec=object)
            if self.parent is not None:
                self.parent().attach_mock(self._mock, self.name)
        return self._mock


class ConfinedModel(BaseModel):
    """A base class confined to explicitly defined fields in the model schema."""

    model_config = ConfigDict(
        extra="forbid",
    )


def error_if_none(value: T | None, msg: str) -> T:
    """Check and return the value if not None.

    :param value: The value to check
    :param msg: The `RuntimeError` message to raise if it is None
    :raises RuntimeError: If the value is None
    :returns: The value if not None

    Used to implement a pattern where a variable is None at init, then
    changed by a method, then used in a later method.
    """
    if value is None:
        raise RuntimeError(msg)
    return value



import asyncio

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    DetectorController,
    DetectorTrigger,
    TriggerInfo,
    wait_for_value,
)

from ._eiger_io import EigerDriverIO, EigerTriggerMode

EIGER_TRIGGER_MODE_MAP = {
    DetectorTrigger.INTERNAL: EigerTriggerMode.INTERNAL,
    DetectorTrigger.CONSTANT_GATE: EigerTriggerMode.GATE,
    DetectorTrigger.VARIABLE_GATE: EigerTriggerMode.GATE,
    DetectorTrigger.EDGE_TRIGGER: EigerTriggerMode.EDGE,
}


class EigerController(DetectorController):
    def __init__(
        self,
        driver: EigerDriverIO,
    ) -> None:
        self._drv = driver

    def get_deadtime(self, exposure: float | None) -> float:
        # See https://media.dectris.com/filer_public/30/14/3014704e-5f3b-43ba-8ccf-8ef720e60d2a/240202_usermanual_eiger2.pdf
        return 0.0001

    async def set_energy(self, energy: float, tolerance: float = 0.1):
        """Set photon energy."""
        """Changing photon energy takes some time so only do so if the current energy is
        outside the tolerance."""

        current_energy = await self._drv.detector.photon_energy.get_value()
        if abs(current_energy - energy) > tolerance:
            await self._drv.detector.photon_energy.set(energy)

    async def prepare(self, trigger_info: TriggerInfo):
        coros = [
            self._drv.detector.trigger_mode.set(
                EIGER_TRIGGER_MODE_MAP[trigger_info.trigger].value
            ),
            self._drv.detector.nimages.set(trigger_info.total_number_of_exposures),
        ]
        if trigger_info.livetime is not None:
            coros.extend(
                [
                    self._drv.detector.count_time.set(trigger_info.livetime),
                    self._drv.detector.frame_time.set(trigger_info.livetime),
                ]
            )

        await asyncio.gather(*coros)

    async def arm(self):
        # NOTE: This will return immedietly on FastCS 0.8.0,
        # but will return after the Eiger has completed arming in 0.9.0.
        # https://github.com/DiamondLightSource/FastCS/pull/141
        await self._drv.detector.arm.trigger(timeout=DEFAULT_TIMEOUT)

    async def wait_for_idle(self):
        await wait_for_value(self._drv.detector.state, "idle", timeout=DEFAULT_TIMEOUT)

    async def disarm(self):
        await self._drv.detector.disarm.trigger()



from ._utils import StrictEnum


class OnOff(StrictEnum):
    ON = "On"
    OFF = "Off"


class EnableDisable(StrictEnum):
    ENABLE = "Enable"
    DISABLE = "Disable"


class EnabledDisabled(StrictEnum):
    ENABLED = "Enabled"
    DISABLED = "Disabled"


class InOut(StrictEnum):
    IN = "In"
    OUT = "Out"


class YesNo(StrictEnum):
    YES = "Yes"
    NO = "No"



import asyncio
import math
import re
from unittest.mock import ANY, call

import pytest
from bluesky.protocols import Reading

from ophyd_async.core import (
    DerivedSignalFactory,
    SignalRW,
    Table,
    derived_signal_rw,
    soft_signal_rw,
)
from ophyd_async.sim import (
    HorizontalMirror,
    HorizontalMirrorDerived,
    TwoJackDerived,
    TwoJackTransform,
    VerticalMirror,
)
from ophyd_async.testing import (
    assert_describe_signal,
    assert_reading,
    assert_value,
    get_mock,
    set_mock_value,
)


@pytest.mark.parametrize(
    "x1, x2, x, roll",
    [
        (0, 0, 0, 0),
        (0, 1, 0.5, math.pi / 4),
        (2, 1, 1.5, -math.pi / 4),
    ],
)
async def test_get_returns_right_position(x1: float, x2: float, x: float, roll: float):
    inst = HorizontalMirror("mirror")
    await inst.x1.set(x1)
    await inst.x2.set(x2)
    assert inst.x.name == "mirror-x"
    assert inst.roll.name == "mirror-roll"
    for sig, value in [(inst.x, x), (inst.roll, roll)]:
        await assert_value(sig, value)
        await assert_reading(sig, {sig.name: {"value": value}})
        await assert_describe_signal(sig, dtype="number", dtype_numpy="<f8", shape=[])
        location = await sig.locate()
        assert location == {"setpoint": value, "readback": value}


async def assert_mirror_readings(
    results: asyncio.Queue[dict[str, Reading[float]]], x: float, roll: float
):
    for name, value in [("mirror-x", x), ("mirror-roll", roll)]:
        reading = await results.get()
        assert reading == {
            name: {"value": value, "timestamp": ANY, "alarm_severity": 0}
        }
    assert results.empty()


async def test_monitoring_position():
    results = asyncio.Queue[dict[str, Reading[float]]]()
    inst = HorizontalMirror("mirror")
    inst.x.subscribe_reading(results.put_nowait)
    inst.roll.subscribe_reading(results.put_nowait)
    await assert_mirror_readings(results, 0, 0)
    await inst.x2.set(1)
    await assert_mirror_readings(results, 0.5, math.pi / 4)
    inst.x.clear_sub(results.put_nowait)
    inst.roll.clear_sub(results.put_nowait)
    await inst.x1.set(1)
    assert results.empty()


async def test_setting_position_straight_through():
    inst = VerticalMirror("mirror")
    # Connect in mock mode so we can see what would have been set
    await inst.connect(mock=True)
    m = get_mock(inst)
    await inst.set(TwoJackDerived(height=1.5, angle=-math.pi / 4))
    assert m.mock_calls == [
        call.y1.user_setpoint.put(2.0, wait=True),
        call.y2.user_setpoint.put(1.0, wait=True),
    ]
    m.reset_mock()
    # Try to move just one axis
    await inst.height.set(0.5)
    assert m.mock_calls == [
        call.height.put(0.5, wait=True),
        call.y1.user_setpoint.put(1.0, wait=True),
        call.y2.user_setpoint.put(pytest.approx(0.0), wait=True),
    ]
    m.reset_mock()


async def test_setting_position_extra_indirection():
    inst = HorizontalMirror("mirror")
    # Connect in mock mode so we can see what would have been set
    await inst.connect(mock=True)
    m = get_mock(inst)
    await inst.set(HorizontalMirrorDerived(x=1.5, roll=-math.pi / 4))
    assert m.mock_calls == [
        call.x1.user_setpoint.put(2.0, wait=True),
        call.x2.user_setpoint.put(1.0, wait=True),
    ]
    m.reset_mock()
    # Try to move just one axis
    await inst.x.set(0.5)
    assert m.mock_calls == [
        call.x.put(0.5, wait=True),
        call.x1.user_setpoint.put(1.0, wait=True),
        call.x2.user_setpoint.put(pytest.approx(0.0), wait=True),
    ]
    m.reset_mock()


def test_mismatching_args():
    with pytest.raises(
        TypeError,
        match=re.escape(
            "Expected the following to be passed as keyword arguments"
            " {'distance': <class 'float'>, 'jack1': <class 'float'>, "
            "'jack2': <class 'float'>}, "
            "got {'jack1': <class 'float'>, 'jack22': <class 'float'>, "
            "'distance': <class 'float'>}"
        ),
    ):
        DerivedSignalFactory(
            TwoJackTransform,
            jack1=soft_signal_rw(float),
            jack22=soft_signal_rw(float),
            distance=soft_signal_rw(float),
        )


@pytest.fixture
def derived_signal() -> SignalRW[float]:
    signal_r = soft_signal_rw(int, initial_value=4)

    def _get(ts: int) -> float:
        return ts

    async def _put(value: float) -> None:
        pass

    return derived_signal_rw(_get, _put, ts=signal_r)


async def test_derived_signal_backend_connect_pass(
    derived_signal: SignalRW,
):
    result = await derived_signal.connect()
    assert result is None


async def test_derived_signal_backend_set_value(
    derived_signal: SignalRW,
) -> None:
    await derived_signal.connect(mock=True)
    with pytest.raises(RuntimeError):
        set_mock_value(derived_signal, 1.0)


async def test_derived_signal_backend_put_wait_fails(
    derived_signal: SignalRW,
) -> None:
    with pytest.raises(RuntimeError):
        await derived_signal.set(value=None, wait=False)
    with pytest.raises(RuntimeError):
        await derived_signal.set(value=None, wait=True)


def test_make_rw_signal_type_mismatch():
    factory = DerivedSignalFactory(
        TwoJackTransform,
        set_derived=None,
        distance=soft_signal_rw(float),
        jack1=soft_signal_rw(float),
        jack2=soft_signal_rw(float),
    )
    with pytest.raises(
        ValueError,
        match=re.escape("Must define a set_derived method to support derived"),
    ):
        factory.derived_signal_rw(datatype=Table, name="")



"""Demo EPICS Devices for the tutorial."""

from ._ioc import start_ioc_subprocess
from ._motor import DemoMotor
from ._point_detector import DemoPointDetector
from ._point_detector_channel import DemoPointDetectorChannel, EnergyMode
from ._stage import DemoStage

__all__ = [
    "DemoMotor",
    "DemoStage",
    "EnergyMode",
    "DemoPointDetectorChannel",
    "DemoPointDetector",
    "start_ioc_subprocess",
]



"""Support for the ADAndor areaDetector driver.

https://github.com/areaDetector/ADAndor.
"""

from ._andor import Andor2Detector
from ._andor_controller import Andor2Controller
from ._andor_io import Andor2DriverIO, Andor2TriggerMode

__all__ = [
    "Andor2Detector",
    "Andor2Controller",
    "Andor2DriverIO",
    "Andor2TriggerMode",
]



import re

from autodoc2.render.myst_ import MystRenderer

DOTTED_MODULE = re.compile(r"((?:[a-zA-Z_][a-zA-Z_0-9]+\.)+)")


class ShortenedNamesRenderer(MystRenderer):
    def format_annotation(self, annotation):
        if annotation:
            annotation = DOTTED_MODULE.sub(r"~\1", annotation)
        return super().format_annotation(annotation)



"""Support for the ADAravis areaDetector driver.

https://github.com/areaDetector/ADAravis
"""

from ._aravis import AravisDetector
from ._aravis_controller import AravisController
from ._aravis_io import AravisDriverIO, AravisTriggerSource

__all__ = [
    "AravisDetector",
    "AravisController",
    "AravisDriverIO",
    "AravisTriggerSource",
]



import asyncio
import time
from collections import defaultdict
from collections.abc import AsyncGenerator, AsyncIterator, Sequence
from typing import Any
from unittest.mock import Mock

import bluesky.plan_stubs as bps
import pytest
from bluesky.protocols import StreamAsset
from bluesky.run_engine import RunEngine
from event_model import (  # type: ignore
    ComposeStreamResourceBundle,
    DataKey,
    compose_stream_resource,
)
from pydantic import ValidationError

from ophyd_async.core import (
    DetectorController,
    DetectorTrigger,
    DetectorWriter,
    FlyerController,
    StandardDetector,
    StandardFlyer,
    StrictEnum,
    TriggerInfo,
    observe_value,
)
from ophyd_async.epics.core import epics_signal_rw
from ophyd_async.testing import assert_emitted


class TriggerState(StrictEnum):
    NULL = "null"
    PREPARING = "preparing"
    STARTING = "starting"
    STOPPING = "stopping"


class DummyTriggerLogic(FlyerController[int]):
    def __init__(self):
        self.state = TriggerState.NULL

    async def prepare(self, value: int):
        self.state = TriggerState.PREPARING
        return value

    async def kickoff(self):
        self.state = TriggerState.STARTING

    async def complete(self):
        self.state = TriggerState.NULL

    async def stop(self):
        self.state = TriggerState.STOPPING


class DummyWriter(DetectorWriter):
    def __init__(self, name: str, shape: Sequence[int]):
        self.dummy_signal = epics_signal_rw(int, "pva://read_pv")
        self._shape = shape
        self._name = name
        self._file: ComposeStreamResourceBundle | None = None
        self._last_emitted = 0
        self.index = 0

    async def open(self, name: str, exposures_per_event: int = 1) -> dict[str, DataKey]:
        self._exposures_per_event = exposures_per_event
        return {
            self._name: DataKey(
                source="soft://some-source",
                shape=[exposures_per_event, *self._shape],
                dtype="number",
                dtype_numpy="<u2",
                external="STREAM:",
            )
        }

    async def observe_indices_written(
        self, timeout: float
    ) -> AsyncGenerator[int, None]:
        num_captured: int
        async for num_captured in observe_value(self.dummy_signal, timeout):
            yield num_captured // self._exposures_per_event

    async def get_indices_written(self) -> int:
        return self.index // self._exposures_per_event

    async def collect_stream_docs(
        self, name: str, indices_written: int
    ) -> AsyncIterator[StreamAsset]:
        if indices_written:
            if not self._file:
                self._file = compose_stream_resource(
                    mimetype="application/x-hdf5",
                    uri="file://",
                    data_key=self._name,
                    parameters={
                        "path": "",
                        "dataset": "",
                    },
                    uid=None,
                    validate=True,
                )
                yield "stream_resource", self._file.stream_resource_doc

            if indices_written >= self._last_emitted:
                indices = {
                    "start": self._last_emitted,
                    "stop": indices_written,
                }
                self._last_emitted = indices_written
                self._last_flush = time.monotonic()
                yield "stream_datum", self._file.compose_stream_datum(indices)

    async def close(self) -> None:
        self._file = None


@pytest.fixture
async def detectors(RE: RunEngine) -> tuple[StandardDetector, StandardDetector]:
    writers = [DummyWriter("testa", (1, 1)), DummyWriter("testb", (1, 1))]
    await writers[0].dummy_signal.connect(mock=True)
    await writers[1].dummy_signal.connect(mock=True)

    def dummy_arm_1(self=None, trigger=None, num=0, exposure=None):
        return writers[0].dummy_signal.set(1)

    async def dummy_arm_2(self=None, trigger=None, num=0, exposure=None):
        return writers[1].dummy_signal.set(1)

    detector_1: StandardDetector[Any, Any] = StandardDetector(
        Mock(spec=DetectorController, get_deadtime=lambda num: num, arm=dummy_arm_1),
        writers[0],
        name="detector_1",
    )
    detector_2: StandardDetector[Any, Any] = StandardDetector(
        Mock(spec=DetectorController, get_deadtime=lambda num: num, arm=dummy_arm_2),
        writers[1],
        name="detector_2",
    )

    return (detector_1, detector_2)


@pytest.mark.parametrize(
    "number_of_triggers", [[1, 2, 3, 4], [2, 3, 100, 3], [1, 1, 1, 1]]
)
async def test_hardware_triggered_flyable(
    RE: RunEngine, detectors: tuple[StandardDetector], number_of_triggers: list[int]
):
    docs = defaultdict(list)
    RE.subscribe(lambda name, doc: docs[name].append(doc))

    trigger_logic = DummyTriggerLogic()
    flyer = StandardFlyer(trigger_logic, name="flyer")

    def flying_plan():
        yield from bps.stage_all(*detectors, flyer)
        assert flyer._trigger_logic.state == TriggerState.STOPPING

        # move the flyer to the correct place, before fly scanning.
        # Prepare the flyer first to get the trigger info for the detectors
        yield from bps.prepare(flyer, 1, wait=True)

        # prepare detectors second.
        for detector in detectors:
            yield from bps.prepare(
                detector,
                TriggerInfo(
                    number_of_events=number_of_triggers,
                    trigger=DetectorTrigger.CONSTANT_GATE,
                    deadtime=2,
                    livetime=2,
                ),
                wait=True,
            )

        assert flyer._trigger_logic.state == TriggerState.PREPARING
        for detector in detectors:
            detector._controller.disarm.assert_called_once()  # type: ignore

        yield from bps.open_run()
        yield from bps.declare_stream(*detectors, name="main_stream", collect=True)
        frames_completed: int = 0
        for frames in number_of_triggers:
            yield from bps.kickoff(flyer)
            for detector in detectors:
                yield from bps.kickoff(detector)

            yield from bps.complete(flyer, wait=False, group="complete")
            for detector in detectors:
                yield from bps.complete(detector, wait=False, group="complete")

            assert flyer._trigger_logic.state == TriggerState.NULL

            # Manually increment the index as if a frame was taken
            frames_completed += frames
            for detector in detectors:
                yield from bps.abs_set(detector._writer.dummy_signal, frames_completed)
                detector._writer.index = frames_completed
            done = False
            while not done:
                try:
                    yield from bps.wait(group="complete", timeout=0.5)
                except TimeoutError:
                    pass
                else:
                    done = True

                    yield from bps.collect(
                        *detectors,
                        return_payload=False,
                        name="main_stream",
                    )
            yield from bps.wait(group="complete")

        for detector in detectors:
            # ensure _completable_frames are reset after completion
            assert detector._completable_exposures == 0

        yield from bps.close_run()

        yield from bps.unstage_all(flyer, *detectors)
        for detector in detectors:
            assert detector._controller.disarm.called  # type: ignore
        assert trigger_logic.state == TriggerState.STOPPING

    # fly scan
    RE(flying_plan())

    assert_emitted(
        docs,
        start=1,
        descriptor=1,
        stream_resource=2,
        stream_datum=2 * len(number_of_triggers),
        stop=1,
    )
    # test stream datum
    seq_numbers: list = []
    frame_completed: int = 0
    last_frame_collected: int = 0
    for frame in number_of_triggers:
        frame_completed += frame
        seq_numbers.extend([(last_frame_collected, frame_completed)] * 2)
        last_frame_collected = frame_completed
    for index, stream_datum in enumerate(docs["stream_datum"]):
        assert stream_datum["descriptor"] == docs["descriptor"][0]["uid"]
        assert stream_datum["seq_nums"] == {
            "start": seq_numbers[index][0] + 1,
            "stop": seq_numbers[index][1] + 1,
        }
        assert stream_datum["indices"] == {
            "start": seq_numbers[index][0],
            "stop": seq_numbers[index][1],
        }
        assert stream_datum["stream_resource"] in [
            sd["uid"].split("/")[0] for sd in docs["stream_datum"]
        ]


@pytest.mark.timeout(3.0)
@pytest.mark.parametrize(
    "number_of_triggers,invoke_extra_kickoff_before_complete",
    [
        (10, True),
        ([10], True),
        (10, False),
        ([10], False),
    ],
)
async def test_hardware_triggered_flyable_too_many_kickoffs(
    RE: RunEngine,
    detectors: tuple[StandardDetector],
    number_of_triggers: int | list[int],
    invoke_extra_kickoff_before_complete: bool,
):
    trigger_logic = DummyTriggerLogic()
    flyer = StandardFlyer(trigger_logic, name="flyer")
    trigger_info = TriggerInfo(
        number_of_events=number_of_triggers,
        trigger=DetectorTrigger.CONSTANT_GATE,
        deadtime=2,
        livetime=2,
    )

    def flying_plan():
        yield from bps.stage_all(*detectors, flyer)
        assert flyer._trigger_logic.state == TriggerState.STOPPING

        # move the flyer to the correct place, before fly scanning.
        # Prepare the flyer first to get the trigger info for the detectors
        yield from bps.prepare(flyer, 1, wait=True)

        # prepare detectors second.
        for detector in detectors:
            yield from bps.prepare(
                detector,
                trigger_info,
                wait=True,
            )

        yield from bps.open_run()
        yield from bps.declare_stream(*detectors, name="main_stream", collect=True)

        yield from bps.kickoff(flyer)
        for detector in detectors:
            yield from bps.kickoff(detector)
            # Perform an additional kickoff
            if invoke_extra_kickoff_before_complete:
                yield from bps.kickoff(detector)
        yield from bps.complete(flyer, wait=False, group="complete")
        for detector in detectors:
            yield from bps.complete(detector, wait=False, group="complete")

        assert flyer._trigger_logic.state == TriggerState.NULL

        # Manually increment the index as if a frame was taken
        for detector in detectors:
            yield from bps.abs_set(
                detector._writer.dummy_signal, trigger_info.total_number_of_exposures
            )
            detector._writer.index = trigger_info.total_number_of_exposures

        yield from bps.wait(group="complete")

        yield from bps.collect(
            *detectors,
            return_payload=False,
            name="main_stream",
        )

        for detector in detectors:
            # Since we set number of iterations to 1 (default),
            # make sure it gets reset on complete
            assert detector._completable_exposures == 0
            assert detector._events_to_complete == 0
            assert detector._number_of_events_iter is None
            assert detector._controller.wait_for_idle.called  # type: ignore

            # This is an additional kickoff
            # Ensuring stop iteration is called if kickoff is invoked after complete
            if not invoke_extra_kickoff_before_complete:
                yield from bps.kickoff(detector)
        yield from bps.close_run()

        yield from bps.unstage_all(flyer, *detectors)

    # fly scan
    if invoke_extra_kickoff_before_complete:
        match_msg = "Kickoff called more than the configured number"
    else:
        match_msg = "Prepare must be called before kickoff!"
    with pytest.raises(Exception, match=match_msg):
        RE(flying_plan())

    # Try explicitly letting event loop clean up tasks...?
    await asyncio.sleep(1.0)


@pytest.mark.parametrize(
    ["kwargs", "error_msg"],
    [
        (
            {
                "number_of_events": 1,
                "trigger": DetectorTrigger.CONSTANT_GATE,
                "deadtime": 2,
                "livetime": 2,
                "exposure_timeout": "a",
            },
            "Input should be a valid number, unable to parse string as a number "
            "[type=float_parsing, input_value='a', input_type=str]",
        ),
        (
            {
                "number_of_events": 1,
                "trigger": "CONSTANT_GATE",
                "deadtime": 2,
                "livetime": -1,
            },
            "Input should be greater than or equal to 0 "
            "[type=greater_than_equal, input_value=-1, input_type=int]",
        ),
        (
            {
                "number_of_events": 1,
                "trigger": DetectorTrigger.INTERNAL,
                "deadtime": 2,
                "livetime": 1,
                "exposure_timeout": -1,
            },
            "Input should be greater than 0 "
            "[type=greater_than, input_value=-1, input_type=int]",
        ),
        (
            {
                "number_of_events": 1,
                "trigger": "not_in_enum",
                "deadtime": 2,
                "livetime": 1,
                "exposure_timeout": None,
            },
            "Input should be 'INTERNAL', 'EDGE_TRIGGER', 'CONSTANT_GATE' or "
            "'VARIABLE_GATE' [type=enum, input_value='not_in_enum', input_type=str]",
        ),
        (
            {
                "number_of_events": -100,
                "trigger": "CONSTANT_GATE",
                "deadtime": 2,
                "livetime": 1,
            },
            "number_of_events.constrained-int\n  Input should be greater than or "
            "equal to 0 [type=greater_than_equal, input_value=-100, input_type=int]",
        ),
        (
            {
                "number_of_events": [1, 1, 1, 1, -100],
                "trigger": "CONSTANT_GATE",
                "deadtime": 2,
                "livetime": 1,
            },
            "number_of_events.list[constrained-int].4\n"
            "  Input should be greater than or equal to 0 [type=greater_than_equal,"
            " input_value=-100, input_type=int]\n",
        ),
    ],
)
def test_malformed_trigger_info(kwargs, error_msg):
    with pytest.raises(ValidationError) as exc:
        TriggerInfo(**kwargs)
    assert error_msg in str(exc.value)



import asyncio
from collections.abc import AsyncIterator
from typing import TypeGuard
from xml.etree import ElementTree as ET

from bluesky.protocols import StreamAsset
from event_model import DataKey
from pydantic import PositiveInt

from ophyd_async.core import (
    DatasetDescriber,
    HDFDatasetDescription,
    HDFDocumentComposer,
    PathProvider,
)

from ._core_io import NDFileHDFIO, NDPluginBaseIO
from ._core_writer import ADWriter
from ._utils import (
    convert_param_dtype_to_np,
    convert_pv_dtype_to_np,
)


def _is_fully_described(shape: tuple[int | None, ...]) -> TypeGuard[tuple[int, ...]]:
    return None not in shape


class ADHDFWriter(ADWriter[NDFileHDFIO]):
    """Allow `NDFileHDFIO` to be used within `StandardDetector`."""

    default_suffix: str = "HDF1:"

    def __init__(
        self,
        fileio: NDFileHDFIO,
        path_provider: PathProvider,
        dataset_describer: DatasetDescriber,
        plugins: dict[str, NDPluginBaseIO] | None = None,
    ) -> None:
        super().__init__(
            fileio,
            path_provider,
            dataset_describer,
            plugins=plugins,
            file_extension=".h5",
            mimetype="application/x-hdf5",
        )
        self._datasets: list[HDFDatasetDescription] = []
        self._composer: HDFDocumentComposer | None = None
        self._filename_template = "%s%s"

    async def open(
        self, name: str, exposures_per_event: PositiveInt = 1
    ) -> dict[str, DataKey]:
        # Setting HDF writer specific signals
        # Make sure we are using chunk auto-sizing
        await asyncio.gather(self.fileio.chunk_size_auto.set(True))

        await asyncio.gather(
            self.fileio.num_extra_dims.set(0),
            self.fileio.lazy_open.set(True),
            self.fileio.swmr_mode.set(True),
            self.fileio.xml_file_name.set(""),
        )

        self._path_info = self._path_provider(device_name=name)

        # Set common AD file plugin params, begin capturing
        await self._begin_capture(name)

        detector_shape = await self._dataset_describer.shape()
        np_dtype = await self._dataset_describer.np_datatype()

        # Used by the base class
        self._exposures_per_event = exposures_per_event

        # Determine number of frames that will be saved per HDF chunk.
        # On a fresh IOC startup, this is set to zero until the first capture,
        # so if it is zero, set it to 1.
        frames_per_chunk = await self.fileio.num_frames_chunks.get_value()
        if frames_per_chunk == 0:
            frames_per_chunk = 1
            await self.fileio.num_frames_chunks.set(frames_per_chunk)

        if not _is_fully_described(detector_shape):
            # Questions:
            # - Can AreaDetector support this?
            # - How to deal with chunking?
            # Don't support for now - leave option open to support it later
            raise ValueError(
                "Datasets with partially unknown dimensionality "
                "are not currently supported by ADHDFWriter."
            )

        # Add the main data
        self._datasets = [
            HDFDatasetDescription(
                data_key=name,
                dataset="/entry/data/data",
                shape=(exposures_per_event, *detector_shape),
                dtype_numpy=np_dtype,
                chunk_shape=(frames_per_chunk, *detector_shape),
            )
        ]

        # And all the scalar datasets
        for plugin in self._plugins.values():
            maybe_xml = await plugin.nd_attributes_file.get_value()
            # This is the check that ADCore does to see if it is an XML string
            # rather than a filename to parse
            if "<Attributes>" in maybe_xml:
                root = ET.fromstring(maybe_xml)
                for child in root:
                    data_key = child.attrib["name"]
                    if child.attrib.get("type", "EPICS_PV") == "EPICS_PV":
                        np_datatype = convert_pv_dtype_to_np(
                            child.attrib.get("dbrtype", "DBR_NATIVE")
                        )
                    else:
                        np_datatype = convert_param_dtype_to_np(
                            child.attrib.get("datatype", "INT")
                        )
                    self._datasets.append(
                        HDFDatasetDescription(
                            data_key=data_key,
                            dataset=f"/entry/instrument/NDAttributes/{data_key}",
                            shape=(exposures_per_event,)
                            if exposures_per_event > 1
                            else (),
                            dtype_numpy=np_datatype,
                            # NDAttributes appear to always be configured with
                            # this chunk size
                            chunk_shape=(16384,),
                        )
                    )

        self._composer = HDFDocumentComposer(
            # See https://github.com/bluesky/ophyd-async/issues/122
            f"{self._path_info.directory_uri}{self._path_info.filename}{self._file_extension}",
            self._datasets,
        )

        describe = {
            ds.data_key: DataKey(
                source=self.fileio.full_file_name.source,
                shape=list(ds.shape),
                dtype="array"
                if exposures_per_event > 1 or len(ds.shape) > 1
                else "number",
                dtype_numpy=ds.dtype_numpy,
                external="STREAM:",
            )
            for ds in self._datasets
        }
        return describe

    async def collect_stream_docs(
        self, name: str, indices_written: int
    ) -> AsyncIterator[StreamAsset]:
        # TODO: fail if we get dropped frames
        if self._composer is None:
            msg = f"open() not called on {self}"
            raise RuntimeError(msg)
        await self.fileio.flush_now.set(True)
        for doc in self._composer.make_stream_docs(indices_written):
            yield doc



from ophyd_async.core import (
    AsyncStatus,
    PathProvider,
    StandardDetector,
    TriggerInfo,
)
from ophyd_async.epics.odin import Odin, OdinWriter

from ._eiger_controller import EigerController
from ._eiger_io import EigerDriverIO


class EigerDetector(StandardDetector):
    """Ophyd-async implementation of an Eiger Detector."""

    _controller: EigerController
    _writer: OdinWriter

    def __init__(
        self,
        prefix: str,
        path_provider: PathProvider,
        drv_suffix="-EA-EIGER-01:",
        hdf_suffix="-EA-EIGER-01:OD:",
        odin_nodes: int = 4,
        name="",
    ):
        self.drv = EigerDriverIO(prefix + drv_suffix)
        self.odin = Odin(prefix + hdf_suffix, nodes=odin_nodes)

        super().__init__(
            EigerController(self.drv),
            OdinWriter(
                path_provider,
                self.odin,
                self.drv.detector.bit_depth_image,
            ),
            name=name,
        )

    @AsyncStatus.wrap
    async def prepare(self, value: TriggerInfo) -> None:
        await super().prepare(value)



from collections.abc import Sequence
from dataclasses import dataclass
from xml.etree import ElementTree as ET

from ophyd_async.core import (
    SignalR,
    StrictEnum,
    SubsetEnum,
    SupersetEnum,
)


class ADBaseDataType(SupersetEnum):
    INT8 = "Int8"
    UINT8 = "UInt8"
    INT16 = "Int16"
    UINT16 = "UInt16"
    INT32 = "Int32"
    UINT32 = "UInt32"
    INT64 = "Int64"
    UINT64 = "UInt64"
    FLOAT32 = "Float32"
    FLOAT64 = "Float64"
    # Driver database override will blank the enum string if it doesn't
    # support a datatype
    UNDEFINED = ""


def convert_ad_dtype_to_np(ad_dtype: ADBaseDataType) -> str:
    ad_dtype_to_np_dtype = {
        ADBaseDataType.INT8: "|i1",
        ADBaseDataType.UINT8: "|u1",
        ADBaseDataType.INT16: "<i2",
        ADBaseDataType.UINT16: "<u2",
        ADBaseDataType.INT32: "<i4",
        ADBaseDataType.UINT32: "<u4",
        ADBaseDataType.INT64: "<i8",
        ADBaseDataType.UINT64: "<u8",
        ADBaseDataType.FLOAT32: "<f4",
        ADBaseDataType.FLOAT64: "<f8",
    }
    np_type = ad_dtype_to_np_dtype.get(ad_dtype)
    if np_type is None:
        raise ValueError(
            "Areadetector driver has a blank DataType, this is not supported"
        )
    return np_type


def convert_pv_dtype_to_np(datatype: str) -> str:
    _pvattribute_to_ad_datatype = {
        "DBR_SHORT": ADBaseDataType.INT16,
        "DBR_ENUM": ADBaseDataType.INT16,
        "DBR_INT": ADBaseDataType.INT32,
        "DBR_LONG": ADBaseDataType.INT32,
        "DBR_FLOAT": ADBaseDataType.FLOAT32,
        "DBR_DOUBLE": ADBaseDataType.FLOAT64,
    }
    if datatype in ["DBR_STRING", "DBR_CHAR"]:
        np_datatype = "s40"
    elif datatype == "DBR_NATIVE":
        raise ValueError("Don't support DBR_NATIVE yet")
    else:
        try:
            np_datatype = convert_ad_dtype_to_np(_pvattribute_to_ad_datatype[datatype])
        except KeyError as exc:
            raise ValueError(f"Invalid dbr type {datatype}") from exc
    return np_datatype


def convert_param_dtype_to_np(datatype: str) -> str:
    _paramattribute_to_ad_datatype = {
        "INT": ADBaseDataType.INT32,
        "INT64": ADBaseDataType.INT64,
        "DOUBLE": ADBaseDataType.FLOAT64,
    }
    if datatype in ["STRING"]:
        np_datatype = "s40"
    else:
        try:
            np_datatype = convert_ad_dtype_to_np(
                _paramattribute_to_ad_datatype[datatype]
            )
        except KeyError as exc:
            raise ValueError(f"Invalid datatype {datatype}") from exc
    return np_datatype


class ADFileWriteMode(StrictEnum):
    SINGLE = "Single"
    CAPTURE = "Capture"
    STREAM = "Stream"


class ADImageMode(SubsetEnum):
    SINGLE = "Single"
    MULTIPLE = "Multiple"
    CONTINUOUS = "Continuous"


class NDAttributeDataType(StrictEnum):
    INT = "INT"
    DOUBLE = "DOUBLE"
    STRING = "STRING"


class NDAttributePvDbrType(StrictEnum):
    DBR_SHORT = "DBR_SHORT"
    DBR_ENUM = "DBR_ENUM"
    DBR_INT = "DBR_INT"
    DBR_LONG = "DBR_LONG"
    DBR_FLOAT = "DBR_FLOAT"
    DBR_DOUBLE = "DBR_DOUBLE"
    DBR_STRING = "DBR_STRING"
    DBR_CHAR = "DBR_CHAR"


@dataclass
class NDAttributePv:
    name: str  # name of attribute stamped on array, also scientifically useful name
    # when appended to device.name
    signal: SignalR  # caget the pv given by signal.source and attach to each frame
    dbrtype: NDAttributePvDbrType
    description: str = ""  # A description that appears in the HDF file as an attribute


@dataclass
class NDAttributeParam:
    name: str  # name of attribute stamped on array, also scientifically useful name
    # when appended to device.name
    param: str  # The parameter string as seen in the INP link of the record
    datatype: NDAttributeDataType  # The datatype of the parameter
    addr: int = 0  # The address as seen in the INP link of the record
    description: str = ""  # A description that appears in the HDF file as an attribute


def ndattributes_to_xml(
    ndattributes: Sequence[NDAttributeParam | NDAttributePv],
) -> str:
    """Convert a set of NDAttribute params to XML."""
    root = ET.Element("Attributes")
    for ndattribute in ndattributes:
        if isinstance(ndattribute, NDAttributeParam):
            ET.SubElement(
                root,
                "Attribute",
                name=ndattribute.name,
                type="PARAM",
                source=ndattribute.param,
                addr=str(ndattribute.addr),
                datatype=ndattribute.datatype.value,
                description=ndattribute.description,
            )
        elif isinstance(ndattribute, NDAttributePv):
            ET.SubElement(
                root,
                "Attribute",
                name=ndattribute.name,
                type="EPICS_PV",
                source=ndattribute.signal.source.split("ca://")[-1],
                dbrtype=ndattribute.dbrtype.value,
                description=ndattribute.description,
            )
        else:
            raise ValueError(
                f"Invalid type for ndattributes: {type(ndattribute)}. "
                "Expected NDAttributePv or NDAttributeParam."
            )
    xml_text = ET.tostring(root, encoding="unicode")
    return xml_text



from __future__ import annotations

from collections.abc import Sequence

from ophyd_async.core import PathProvider, SignalR, StandardDetector
from ophyd_async.fastcs.core import fastcs_connector

from ._block import CommonPandaBlocks
from ._control import PandaPcapController
from ._writer import PandaHDFWriter

MINIMUM_PANDA_IOC = "0.11.4"


class HDFPanda(
    CommonPandaBlocks, StandardDetector[PandaPcapController, PandaHDFWriter]
):
    """PandA with common blocks for standard HDF writing."""

    def __init__(
        self,
        prefix: str,
        path_provider: PathProvider,
        config_sigs: Sequence[SignalR] = (),
        name: str = "",
    ):
        error_hint = f"Is PandABlocks-ioc at least version {MINIMUM_PANDA_IOC}?"
        # This has to be first so we make self.pcap
        connector = fastcs_connector(self, prefix, error_hint)
        controller = PandaPcapController(pcap=self.pcap)
        writer = PandaHDFWriter(
            path_provider=path_provider,
            panda_data_block=self.data,
        )
        super().__init__(
            controller=controller,
            writer=writer,
            config_sigs=config_sigs,
            name=name,
            connector=connector,
        )



from ._counter import TangoCounter
from ._detector import TangoDetector
from ._mover import TangoMover
from ._tango import DemoCounter, DemoMover

__all__ = [
    "DemoCounter",
    "DemoMover",
    "TangoCounter",
    "TangoMover",
    "TangoDetector",
]



from collections.abc import Sequence

from ophyd_async.core import PathProvider, SignalR
from ophyd_async.epics import adcore

from ._aravis_controller import AravisController
from ._aravis_io import AravisDriverIO


class AravisDetector(adcore.AreaDetector[AravisController]):
    """Implementation of an ADAravis Detector.

    The detector may be configured for an external trigger on a GPIO port,
    which must be done prior to preparing the detector
    """

    def __init__(
        self,
        prefix: str,
        path_provider: PathProvider,
        drv_suffix="cam1:",
        writer_cls: type[adcore.ADWriter] = adcore.ADHDFWriter,
        fileio_suffix: str | None = None,
        name: str = "",
        config_sigs: Sequence[SignalR] = (),
        plugins: dict[str, adcore.NDPluginBaseIO] | None = None,
    ):
        driver = AravisDriverIO(prefix + drv_suffix)
        controller = AravisController(driver)
        writer = writer_cls.with_io(
            prefix,
            path_provider,
            dataset_source=driver,
            fileio_suffix=fileio_suffix,
            plugins=plugins,
        )
        super().__init__(
            controller=controller,
            writer=writer,
            plugins=plugins,
            name=name,
            config_sigs=config_sigs,
        )



import re

import inflection
import pytest

from ophyd_async.core import Device, Signal

# Need to import them all so the subclass walking gets all subclasses
# If we forget then the full test suite will find the subclasses, but
# running just this test will only get the ones at the top of this file
from ophyd_async.epics import (
    adaravis,  # noqa
    adcore,
    adkinetix,  # noqa
    adpilatus,  # noqa
    adsimdetector,  # noqa
    advimba,  # noqa
)


def get_rec_subclasses(cls: type):
    yield cls
    for subcls in cls.__subclasses__():
        yield from get_rec_subclasses(subcls)


@pytest.mark.parametrize("cls", list(get_rec_subclasses(adcore.NDArrayBaseIO)))
async def test_regularly_named_attributes(cls: Device):
    io = cls("")
    for name, device in io.children():
        check_name(name, device)


def check_name(name: str, device: Device):
    if isinstance(device, Signal):
        pv = extract_last_pv_part(device.source)
        # remove trailing underscore from name,
        # used to resolve clashes with Bluesky terms
        name = name[:-1] if name.endswith("_") else name
        assert inflection.underscore(pv) == name
    else:
        for name, signal in device.children():
            check_name(name, signal)


def extract_last_pv_part(raw_pv):
    """Extracts prefices and _RBV suffices.

    e.g. extracts DEVICE from the following
    ca://DEVICE
    ca://SYSTEM:DEVICE
    ca://SYSTEM:DEVICE_RBV
    """
    pattern = re.compile(
        r"""
        ca://           # Literal prefix "ca://"
        (?:.*:)?        # Optional prefix ending with a colon (non-capturing)
        ([^:_]+)        # Capturing group: base name without colon or underscore
        (?:_RBV)?       # Optional "_RBV" suffix (non-capturing)
        $               # End of string
        """,
        re.VERBOSE,
    )

    match = pattern.search(raw_pv)
    return str(match.group(1) if match else None)



from collections.abc import Sequence

import numpy as np
from pydantic import model_validator

from ophyd_async.core import Array1D, StrictEnum, Table


class PandaHdf5DatasetType(StrictEnum):
    """Dataset options for HDF capture."""

    FLOAT_64 = "float64"
    UINT_32 = "uint32"


class DatasetTable(Table):
    name: Sequence[str]
    dtype: Sequence[PandaHdf5DatasetType]


class SeqTrigger(StrictEnum):
    """Trigger options for the SeqTable."""

    IMMEDIATE = "Immediate"
    BITA_0 = "BITA=0"
    BITA_1 = "BITA=1"
    BITB_0 = "BITB=0"
    BITB_1 = "BITB=1"
    BITC_0 = "BITC=0"
    BITC_1 = "BITC=1"
    POSA_GT = "POSA>=POSITION"
    POSA_LT = "POSA<=POSITION"
    POSB_GT = "POSB>=POSITION"
    POSB_LT = "POSB<=POSITION"
    POSC_GT = "POSC>=POSITION"
    POSC_LT = "POSC<=POSITION"


class SeqTable(Table):
    """Data type for the panda seq table."""

    repeats: Array1D[np.uint16]
    trigger: Sequence[SeqTrigger]
    position: Array1D[np.int32]
    time1: Array1D[np.uint32]
    outa1: Array1D[np.bool_]
    outb1: Array1D[np.bool_]
    outc1: Array1D[np.bool_]
    outd1: Array1D[np.bool_]
    oute1: Array1D[np.bool_]
    outf1: Array1D[np.bool_]
    time2: Array1D[np.uint32]
    outa2: Array1D[np.bool_]
    outb2: Array1D[np.bool_]
    outc2: Array1D[np.bool_]
    outd2: Array1D[np.bool_]
    oute2: Array1D[np.bool_]
    outf2: Array1D[np.bool_]

    @staticmethod
    def row(
        *,
        repeats: int = 1,
        trigger: str = SeqTrigger.IMMEDIATE,
        position: int = 0,
        time1: int = 0,
        outa1: bool = False,
        outb1: bool = False,
        outc1: bool = False,
        outd1: bool = False,
        oute1: bool = False,
        outf1: bool = False,
        time2: int = 0,
        outa2: bool = False,
        outb2: bool = False,
        outc2: bool = False,
        outd2: bool = False,
        oute2: bool = False,
        outf2: bool = False,
    ) -> "SeqTable":
        # Let pydantic do the conversions for us
        return SeqTable(**{k: [v] for k, v in locals().items()})  # type: ignore

    @model_validator(mode="after")
    def _validate_max_length(self) -> "SeqTable":
        # Used to check max_length. Unfortunately trying the ``max_length`` arg in
        # the pydantic field doesn't work.
        first_length = len(self)
        max_length = 4096
        if first_length > max_length:
            msg = f"Length {first_length} is too long"
            raise ValueError(msg)
        return self



from ._vimba import VimbaDetector
from ._vimba_controller import VimbaController
from ._vimba_io import (
    VimbaConvertFormat,
    VimbaDriverIO,
    VimbaExposeOutMode,
    VimbaOverlap,
    VimbaTriggerSource,
)

__all__ = [
    "VimbaDetector",
    "VimbaController",
    "VimbaDriverIO",
    "VimbaExposeOutMode",
    "VimbaTriggerSource",
    "VimbaOverlap",
    "VimbaConvertFormat",
]



from __future__ import annotations

from ophyd_async.core import StandardReadable

from ._base_device import TangoDevice


class TangoReadable(TangoDevice, StandardReadable):
    def __init__(
        self,
        trl: str | None = None,
        name: str = "",
        auto_fill_signals: bool = True,
    ) -> None:
        TangoDevice.__init__(self, trl, name=name, auto_fill_signals=auto_fill_signals)



import pytest

from ophyd_async.core import (
    DetectorTrigger,
    TriggerInfo,
)
from ophyd_async.epics import adaravis, adcore


@pytest.fixture
def test_adaravis(ad_standard_det_factory) -> adaravis.AravisDetector:
    return ad_standard_det_factory(
        adaravis.AravisDetector,
        plugins={"stats": adcore.NDPluginStatsIO("stats_prefix")},
    )


async def test_aravis_name(
    test_adaravis: adaravis.AravisDetector,
):
    assert test_adaravis.name == "test_adaravis1"


@pytest.mark.parametrize("exposure_time", [0.0, 0.1, 1.0, 10.0, 100.0])
async def test_deadtime_invariant_with_exposure_time(
    exposure_time: float,
    test_adaravis: adaravis.AravisDetector,
):
    assert test_adaravis._controller.get_deadtime(exposure_time) == 1961e-6


async def test_unsupported_trigger_excepts(test_adaravis: adaravis.AravisDetector):
    with pytest.raises(
        ValueError,
        # str(EnumClass.value) handling changed in Python 3.11
        match="ADAravis does not support (DetectorTrigger.)?VARIABLE_GATE",
    ):
        await test_adaravis.prepare(
            TriggerInfo(
                number_of_events=0,
                trigger=DetectorTrigger.VARIABLE_GATE,
                deadtime=1,
                livetime=1,
                exposure_timeout=3,
            )
        )



"""Back compat."""

import warnings

from .core import *  # noqa: F403

warnings.warn(
    DeprecationWarning(
        "Use `ophyd_async.epics.core` instead of `ophyd_async.epics.signal` and `pvi`"
    ),
    stacklevel=2,
)

__all__ = []



import asyncio
from typing import Annotated as A

from ophyd_async.core import (
    DatasetDescriber,
    DeviceVector,
    EnableDisable,
    SignalR,
    SignalRW,
    StrictEnum,
)
from ophyd_async.epics.core import EpicsDevice, PvSuffix

from ._utils import ADBaseDataType, ADFileWriteMode, ADImageMode, convert_ad_dtype_to_np


class NDArrayBaseIO(EpicsDevice):
    """Class responsible for passing detector data from drivers to pluglins.

    This mirrors the interface provided by ADCore/db/NDArrayBase.template.
    See HTML docs at https://areadetector.github.io/areaDetector/ADCore/NDArray.html
    """

    unique_id: A[SignalR[int], PvSuffix("UniqueId_RBV")]
    nd_attributes_file: A[SignalRW[str], PvSuffix("NDAttributesFile")]
    acquire: A[SignalRW[bool], PvSuffix.rbv("Acquire")]
    array_size_x: A[SignalR[int], PvSuffix("ArraySizeX_RBV")]
    array_size_y: A[SignalR[int], PvSuffix("ArraySizeY_RBV")]
    data_type: A[SignalR[ADBaseDataType], PvSuffix("DataType_RBV")]
    array_counter: A[SignalRW[int], PvSuffix.rbv("ArrayCounter")]
    # There is no _RBV for this one
    wait_for_plugins: A[SignalRW[bool], PvSuffix("WaitForPlugins")]


class ADBaseDatasetDescriber(DatasetDescriber):
    def __init__(self, driver: NDArrayBaseIO) -> None:
        self._driver = driver

    async def np_datatype(self) -> str:
        return convert_ad_dtype_to_np(await self._driver.data_type.get_value())

    async def shape(self) -> tuple[int, int]:
        shape = await asyncio.gather(
            self._driver.array_size_y.get_value(),
            self._driver.array_size_x.get_value(),
        )
        return shape


class NDPluginBaseIO(NDArrayBaseIO):
    """Base class from which plugins are derived.

    This mirrors the interface provided by ADCore/db/NDPluginBase.template.
    See HTML docs at https://areadetector.github.io/areaDetector/ADCore/NDPluginDriver.html
    """

    nd_array_port: A[SignalRW[str], PvSuffix.rbv("NDArrayPort")]
    enable_callbacks: A[SignalRW[EnableDisable], PvSuffix.rbv("EnableCallbacks")]
    nd_array_address: A[SignalRW[int], PvSuffix.rbv("NDArrayAddress")]
    array_size0: A[SignalR[int], PvSuffix("ArraySize0_RBV")]
    array_size1: A[SignalR[int], PvSuffix("ArraySize1_RBV")]
    queue_size: A[SignalRW[int], PvSuffix.rbv("QueueSize")]


class NDPluginStatsIO(NDPluginBaseIO):
    """Plugin for computing statistics from an image or ROI within an image.

    This mirrors the interface provided by ADCore/db/NDStats.template.
    See HTML docs at https://areadetector.github.io/areaDetector/ADCore/NDPluginStats.html
    """

    # Basic statistics
    compute_statistics: A[SignalRW[bool], PvSuffix.rbv("ComputeStatistics")]
    bgd_width: A[SignalRW[int], PvSuffix.rbv("BgdWidth")]
    total: A[SignalR[float], PvSuffix.rbv("Total")]
    # Centroid statistics
    compute_centroid: A[SignalRW[bool], PvSuffix.rbv("ComputeCentroid")]
    centroid_threshold: A[SignalRW[float], PvSuffix.rbv("CentroidThreshold")]
    # X and Y Profiles
    compute_profiles: A[SignalRW[bool], PvSuffix.rbv("ComputeProfiles")]
    profile_size_x: A[SignalR[int], PvSuffix.rbv("ProfileSizeX")]
    profile_size_y: A[SignalR[int], PvSuffix.rbv("ProfileSizeY")]
    cursor_x: A[SignalRW[int], PvSuffix.rbv("CursorX")]
    cursor_y: A[SignalRW[int], PvSuffix.rbv("CursorY")]
    # Array Histogram
    compute_histogram: A[SignalRW[bool], PvSuffix.rbv("ComputeHistogram")]
    hist_size: A[SignalRW[int], PvSuffix.rbv("HistSize")]
    hist_min: A[SignalRW[float], PvSuffix.rbv("HistMin")]
    hist_max: A[SignalRW[float], PvSuffix.rbv("HistMax")]


class NDROIStatIO(NDPluginBaseIO):
    """Plugin for calculating basic statistics for multiple ROIs.

    Each ROI is implemented as an instance of NDROIStatNIO,
    and the collection of ROIs is held as a DeviceVector.

    See HTML docs at https://areadetector.github.io/areaDetector/ADCore/NDPluginROIStat.html
    """

    def __init__(self, prefix, num_channels=8, with_pvi=False, name=""):
        self.channels = DeviceVector(
            {i: NDROIStatNIO(f"{prefix}{i}:") for i in range(1, num_channels + 1)}
        )
        super().__init__(prefix, with_pvi, name)


class NDROIStatNIO(EpicsDevice):
    """Defines the parameters for a single ROI used for statistics calculation.

    Each instance represents a single ROI, with attributes for its position
    (min_x, min_y) and size (size_x, size_y), as well as a name and use status.

    See definition in ADApp/pluginSrc/NDPluginROIStat.h in https://github.com/areaDetector/ADCore.

    Attributes:
        name: The name of the ROI.
        use: Flag indicating whether the ROI is used.
        min_x: The start X-coordinate of the ROI.
        min_y: The start Y-coordinate of the ROI.
        size_x: The width of the ROI.
        size_y: The height of the ROI.
        min_value: Minimum count value in the ROI.
        max_value: Maximum count value in the ROI.
        mean_value: Mean counts value in the ROI.
        total: Total counts in the ROI.
    """

    name_: A[SignalRW[str], PvSuffix("Name")]
    use: A[SignalRW[bool], PvSuffix.rbv("Use")]
    min_x: A[SignalRW[int], PvSuffix.rbv("MinX")]
    min_y: A[SignalRW[int], PvSuffix.rbv("MinY")]
    size_x: A[SignalRW[int], PvSuffix.rbv("SizeX")]
    size_y: A[SignalRW[int], PvSuffix.rbv("SizeY")]
    # stats
    min_value: A[SignalR[float], PvSuffix("MinValue_RBV")]
    max_value: A[SignalR[float], PvSuffix("MaxValue_RBV")]
    mean_value: A[SignalR[float], PvSuffix("MeanValue_RBV")]
    total: A[SignalR[float], PvSuffix("Total_RBV")]


class ADState(StrictEnum):
    """Default set of states of an AreaDetector driver.

    See definition in ADApp/ADSrc/ADDriver.h in https://github.com/areaDetector/ADCore.
    """

    IDLE = "Idle"
    ACQUIRE = "Acquire"
    READOUT = "Readout"
    CORRECT = "Correct"
    SAVING = "Saving"
    ABORTING = "Aborting"
    ERROR = "Error"
    WAITING = "Waiting"
    INITIALIZING = "Initializing"
    DISCONNECTED = "Disconnected"
    ABORTED = "Aborted"


class ADBaseIO(NDArrayBaseIO):
    """Base class from which areaDetector drivers are derived.

    This mirrors the interface provided by ADCore/db/ADBase.template.
    See HTML docs at https://areadetector.github.io/areaDetector/ADCore/ADDriver.html
    """

    acquire_time: A[SignalRW[float], PvSuffix.rbv("AcquireTime")]
    acquire_period: A[SignalRW[float], PvSuffix.rbv("AcquirePeriod")]
    num_images: A[SignalRW[int], PvSuffix.rbv("NumImages")]
    image_mode: A[SignalRW[ADImageMode], PvSuffix.rbv("ImageMode")]
    detector_state: A[SignalR[ADState], PvSuffix("DetectorState_RBV")]


class ADCompression(StrictEnum):
    NONE = "None"
    NBIT = "N-bit"
    SZIP = "szip"
    ZLIB = "zlib"
    BLOSC = "Blosc"
    BSLZ4 = "BSLZ4"
    LZ4 = "LZ4"
    JPEG = "JPEG"


class NDFileIO(NDArrayBaseIO):
    """Base class from which file writing drivers are derived.

    This mirrors the interface provided by ADCore/ADApp/Db/NDFile.template.
    It does not include any plugin-related fields, for that see NDFilePluginIO.
    """

    file_path: A[SignalRW[str], PvSuffix.rbv("FilePath")]
    file_name: A[SignalRW[str], PvSuffix.rbv("FileName")]
    file_path_exists: A[SignalR[bool], PvSuffix("FilePathExists_RBV")]
    file_template: A[SignalRW[str], PvSuffix.rbv("FileTemplate")]
    full_file_name: A[SignalR[str], PvSuffix("FullFileName_RBV")]
    file_number: A[SignalRW[int], PvSuffix("FileNumber")]
    auto_increment: A[SignalRW[bool], PvSuffix("AutoIncrement")]
    file_write_mode: A[SignalRW[ADFileWriteMode], PvSuffix.rbv("FileWriteMode")]
    num_capture: A[SignalRW[int], PvSuffix.rbv("NumCapture")]
    num_captured: A[SignalR[int], PvSuffix("NumCaptured_RBV")]
    capture: A[SignalRW[bool], PvSuffix.rbv("Capture")]
    array_size0: A[SignalR[int], PvSuffix("ArraySize0")]
    array_size1: A[SignalR[int], PvSuffix("ArraySize1")]
    create_directory: A[SignalRW[int], PvSuffix("CreateDirectory")]


class NDFilePluginIO(NDPluginBaseIO, NDFileIO):
    """Base class from which file plugins are derived.

    This mirrors the interface provided by ADCore/db/NDFilePlugin.template.
    See HTML docs at https://areadetector.github.io/areaDetector/ADCore/NDPluginFile.html
    """

    ...


class NDFileHDFIO(NDFilePluginIO):
    """Plugin for storing data in HDF5 file format.

    This mirrors the interface provided by ADCore/db/NDFileHDF5.template.
    See HTML docs at https://areadetector.github.io/areaDetector/ADCore/NDFileHDF5.html
    """

    position_mode: A[SignalRW[bool], PvSuffix.rbv("PositionMode")]
    compression: A[SignalRW[ADCompression], PvSuffix.rbv("Compression")]
    num_extra_dims: A[SignalRW[int], PvSuffix.rbv("NumExtraDims")]
    swmr_mode: A[SignalRW[bool], PvSuffix.rbv("SWMRMode")]
    flush_now: A[SignalRW[bool], PvSuffix("FlushNow")]
    xml_file_name: A[SignalRW[str], PvSuffix.rbv("XMLFileName")]
    num_frames_chunks: A[SignalRW[int], PvSuffix.rbv("NumFramesChunks")]
    chunk_size_auto: A[SignalRW[bool], PvSuffix.rbv("ChunkSizeAuto")]
    lazy_open: A[SignalRW[bool], PvSuffix.rbv("LazyOpen")]


class NDCBFlushOnSoftTrgMode(StrictEnum):
    ON_NEW_IMAGE = "OnNewImage"
    IMMEDIATELY = "Immediately"


class NDPluginCBIO(NDPluginBaseIO):
    pre_count: A[SignalRW[int], PvSuffix.rbv("PreCount")]
    post_count: A[SignalRW[int], PvSuffix.rbv("PostCount")]
    preset_trigger_count: A[SignalRW[int], PvSuffix.rbv("PresetTriggerCount")]
    trigger: A[SignalRW[bool], PvSuffix.rbv("Trigger")]
    capture: A[SignalRW[bool], PvSuffix.rbv("Capture")]
    flush_on_soft_trg: A[
        SignalRW[NDCBFlushOnSoftTrgMode], PvSuffix.rbv("FlushOnSoftTrg")
    ]



from ._epics_connector import EpicsDeviceConnector, PvSuffix
from ._epics_device import EpicsDevice
from ._pvi_connector import PviDeviceConnector
from ._signal import (
    CaSignalBackend,
    PvaSignalBackend,
    epics_signal_r,
    epics_signal_rw,
    epics_signal_rw_rbv,
    epics_signal_w,
    epics_signal_x,
)
from ._util import stop_busy_record

__all__ = [
    "PviDeviceConnector",
    "EpicsDeviceConnector",
    "PvSuffix",
    "EpicsDevice",
    "CaSignalBackend",
    "PvaSignalBackend",
    "epics_signal_r",
    "epics_signal_rw",
    "epics_signal_rw_rbv",
    "epics_signal_w",
    "epics_signal_x",
    "stop_busy_record",
]



from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
from functools import partial
from typing import TypeVar

import numpy as np
import numpy.typing as npt
from numpy import float64
from scanspec.core import Slice
from velocity_profile.velocityprofile import VelocityProfile

from ophyd_async.core import error_if_none
from ophyd_async.epics.motor import Motor
from ophyd_async.epics.pmac._utils import _PmacMotorInfo

MIN_TURNAROUND = 0.002
MIN_INTERVAL = 0.002

IndexType = TypeVar("IndexType", int, slice, npt.NDArray[np.int_], list[int])


class UserProgram(IntEnum):
    COLLECTION_WINDOW = 1  # Period within a collection window
    GAP = 2  # Transition period between collection windows
    END = 8  # Post-scan state


@dataclass
class PVT:
    """Represents a single position, velocity, and time point for multiple motors."""

    position: dict[Motor, float64]
    velocity: dict[Motor, float64]
    time: float64

    @classmethod
    def default(cls, motors: list[Motor]) -> PVT:
        """Initialise a PVT with zeros.

        :param motors: List of motors
        :returns: PVT instance with zero value placeholders
        """
        return cls(
            position={motor: np.float64(0.0) for motor in motors},
            velocity={motor: np.float64(0.0) for motor in motors},
            time=np.float64(0.0),
        )


@dataclass
class Trajectory:
    positions: dict[Motor, npt.NDArray[np.float64]]
    velocities: dict[Motor, npt.NDArray[np.float64]]
    user_programs: npt.NDArray[np.int32]
    durations: npt.NDArray[np.float64]

    def __len__(self) -> int:
        return len(self.user_programs)

    def with_ramp_down(
        self,
        entry_pvt: PVT,
        ramp_down_pos: dict[Motor, np.float64],
        ramp_down_time: float,
        ramp_down_velocity: float,
    ) -> Trajectory:
        # Make room for additional two points to ramp down
        # from a collection window upper to ramp down position
        trajectory_length = len(self.user_programs)
        total_length = trajectory_length + 2
        motors = ramp_down_pos.keys()

        positions = {motor: np.empty(total_length, float) for motor in motors}
        velocities = {motor: np.empty(total_length, float) for motor in motors}
        durations: npt.NDArray[np.float64] = np.empty(total_length, float)
        user_programs: npt.NDArray[np.int32] = np.empty(total_length, int)

        durations[:trajectory_length] = self.durations
        durations[trajectory_length:] = [entry_pvt.time, ramp_down_time]

        user_programs[:trajectory_length] = self.user_programs
        user_programs[trajectory_length:] = [
            UserProgram.COLLECTION_WINDOW,
            UserProgram.END,
        ]

        for motor in ramp_down_pos.keys():
            positions[motor][:trajectory_length] = self.positions[motor]
            positions[motor][trajectory_length:] = [
                entry_pvt.position[motor],
                ramp_down_pos[motor],
            ]
            velocities[motor][:trajectory_length] = self.velocities[motor]
            velocities[motor][trajectory_length:] = [
                entry_pvt.velocity[motor],
                ramp_down_velocity,
            ]

        return Trajectory(
            positions=positions,
            velocities=velocities,
            user_programs=user_programs,
            durations=durations,
        )

    @classmethod
    def from_slice(
        cls,
        slice: Slice[Motor],
        motor_info: _PmacMotorInfo,
        entry_pvt: PVT | None = None,
        ramp_up_time: float | None = None,
    ) -> tuple[Trajectory, PVT]:
        """Parse a trajectory from a slice.

        :param slice: Information about a series of scan frames along a number of axes
        :param motor_info: Instance of _PmacMotorInfo
        :param entry_pvt: PVT entering this trajectory
        :param ramp_up_time: Time required to ramp up to speed
        :returns Trajectory: Data class representing our parsed trajectory
        """
        # List of indices into slice where gaps occur
        gap_indices: list[int] = np.where(slice.gap)[0].tolist()
        motors = slice.axes()

        # Given a ramp up time is provided, we must be at a
        # gap in the trajectory, that we can ramp up through
        if (not gap_indices or gap_indices[0] != 0) and ramp_up_time:
            raise RuntimeError(
                "Slice must start with a gap, if ramp up time provided. "
                f"Ramp up time given: {ramp_up_time}."
            )

        # Given a ramp up time is provided, we need to construct the PVT
        if (ramp_up_time is None) == (entry_pvt is None):
            raise RuntimeError(
                "Exactly one of ramp_up_time or entry_pvt must be provided."
                f"Provided ramp up time: {ramp_up_time} and entry PVT: {entry_pvt}"
            )

        # Find start and end indices for collection windows
        collection_windows = np.argwhere(
            np.diff(~(slice.gap), prepend=False, append=False)
        ).reshape((-1, 2))

        collection_window_iter = iter(collection_windows)
        sub_traj_funcs = []

        # Given we start at a collection window, insert it
        if not gap_indices or gap_indices[0] != 0:
            start, end = next(collection_window_iter)
            sub_traj_funcs.append(
                partial(
                    Trajectory.from_collection_window,
                    start,
                    end,
                    motors,
                    slice,
                )
            )

        # For each gap, insert a gap, followed by a collection window
        # given the distance to the next gap is greater than 1
        for gap in gap_indices:
            kwargs = {}
            if gap == 0 and ramp_up_time:
                kwargs["ramp_up_time"] = ramp_up_time
            sub_traj_funcs.append(
                partial(
                    Trajectory.from_gap,
                    motor_info,
                    gap,
                    motors,
                    slice,
                    **kwargs,
                )
            )
            if gap != len(slice.gap) - 1 and not slice.gap[gap + 1]:
                start, end = next(collection_window_iter)
                sub_traj_funcs.append(
                    partial(
                        Trajectory.from_collection_window,
                        start,
                        end,
                        motors,
                        slice,
                    )
                )

        sub_trajectories: list[Trajectory] = []
        # If no sub trajectories, initial frame is the end frame
        exit_pvt = entry_pvt
        for func in sub_traj_funcs:
            # Build each sub trajectory, passing the last PVT
            # to the next sub trajectory to build upon
            # Explicitly defining initial and end PVTs
            # to clearly show that the last PVT of a trajectory
            # is passed as the first PVT of the next trajectory
            traj, exit_pvt = func(entry_pvt)
            entry_pvt = exit_pvt
            sub_trajectories.append(traj)

        # Combine sub trajectories
        total_trajectory = Trajectory.from_trajectories(sub_trajectories, motors)

        return total_trajectory, (exit_pvt or PVT.default(motors))

    @classmethod
    def from_trajectories(
        cls, sub_trajectories: list[Trajectory], motors: list[Motor]
    ) -> Trajectory:
        """Parse a trajectory from smaller strajectories.

        :param sub_trajectories: List of trajectories to concatenate
        :returns: Trajectory instance as concatenation of all sub trajectories
        """
        # Initialise arrays to insert sub arrays into
        total_points = sum(len(trajectory) for trajectory in sub_trajectories)
        positions = {motor: np.empty(total_points, float) for motor in motors}
        velocities = {motor: np.empty(total_points, float) for motor in motors}
        durations: npt.NDArray[np.float64] = np.empty(total_points, float)
        user_programs: npt.NDArray[np.int32] = np.empty(total_points, int)

        # Keep track of where we are in overall trajectory
        index_into_trajectory = 0
        for trajectory in sub_trajectories:
            # Insert sub trajectory arrays into overall trajectory arrays
            durations[
                index_into_trajectory : index_into_trajectory + len(trajectory)
            ] = trajectory.durations
            user_programs[
                index_into_trajectory : index_into_trajectory + len(trajectory)
            ] = trajectory.user_programs
            for motor in motors:
                positions[motor][
                    index_into_trajectory : index_into_trajectory + len(trajectory)
                ] = trajectory.positions[motor]
                velocities[motor][
                    index_into_trajectory : index_into_trajectory + len(trajectory)
                ] = trajectory.velocities[motor]
            # Update where we are in the overall trajectory
            index_into_trajectory += len(trajectory)

        trajectory = Trajectory(
            positions=positions,
            velocities=velocities,
            durations=durations,
            user_programs=user_programs,
        )

        return trajectory

    @classmethod
    def from_collection_window(
        cls,
        start: int,
        end: int,
        motors: list[Motor],
        slice: Slice,
        entry_pvt: PVT,
    ) -> tuple[Trajectory, PVT]:
        """Parse a trajectory from a collection window.

        For all frames of the slice that fall within this window, this function will:
          - Insert a sequence of lower → midpoint → lower → ... → midpoint points
            until window ends
          - Calculate and insert 3 point average velocities for these points, using the
            entry PVT to blend with previous trajectories

        :param start: Index into slice where collection window starts
        :param end: Index into slice where collection window end
        :param motors: List of motors involved in trajectory
        :param slice: Information about a series of scan frames along a number of axes
        :param entry_pvt: PVT entering this trajectory
        :returns: Tuple of:
            Trajectory instance encompassing collection window points
            PVT at exit of collection window
        """
        slice_duration = error_if_none(slice.duration, "Slice must have a duration")
        half_durations = slice_duration / 2
        if end > len(half_durations):
            # Clamp collection window if no more frames
            end = len(half_durations)
        trajectory_size = 2 * len(half_durations[start:end])
        positions = {motor: np.empty(trajectory_size, float) for motor in motors}
        velocities = {motor: np.empty(trajectory_size, float) for motor in motors}
        durations: npt.NDArray[np.float64] = np.empty(trajectory_size, float)
        user_programs: npt.NDArray[np.int32] = np.empty(trajectory_size, int)

        # Initialise exit PVT
        exit_pvt = PVT.default(motors)

        for motor in motors:
            # Insert lower -> mid -> lower -> mid... positions
            positions[motor][::2] = slice.lower[motor][start:end]
            positions[motor][1::2] = slice.midpoints[motor][start:end]

            # For velocities we will need the relative distances
            mid_to_upper_velocities = (
                slice.upper[motor][start:end] - slice.midpoints[motor][start:end]
            ) / half_durations[start:end]
            lower_to_mid_velocities = (
                slice.midpoints[motor][start:end] - slice.lower[motor][start:end]
            ) / half_durations[start:end]

            # Smooth first lower point velocity with previous PVT
            velocities[motor][0] = (
                lower_to_mid_velocities[0] + entry_pvt.velocity[motor]
            ) / 2

            # For the midpoints, we take the average of the
            # lower -> mid and mid -> upper velocities of the same point
            velocities[motor][1::2] = (
                lower_to_mid_velocities + mid_to_upper_velocities
            ) / 2

            # For the lower points, we need to take the average of the
            # mid -> upper velocity of the previous point and
            # lower -> mid velocity of the current point
            velocities[motor][2::2] = (
                mid_to_upper_velocities[:-1] + lower_to_mid_velocities[1:]
            ) / 2

            # Exit PVT is the final upper point and its mid to upper velocity
            exit_pvt.position[motor] = slice.upper[motor][end - 1]
            exit_pvt.velocity[motor] = mid_to_upper_velocities[-1]

        durations = np.repeat(
            (half_durations[start:end]),
            2,
        )

        user_programs = np.repeat(UserProgram.COLLECTION_WINDOW, len(user_programs))

        exit_pvt.time = half_durations[end - 1]

        trajectory = Trajectory(
            positions=positions,
            velocities=velocities,
            durations=durations,
            user_programs=user_programs,
        )

        return trajectory, exit_pvt

    @classmethod
    def from_gap(
        cls,
        motor_info: _PmacMotorInfo,
        gap: int,
        motors: list[Motor],
        slice: Slice,
        entry_pvt: PVT,
        ramp_up_time: float | None = None,
    ) -> tuple[Trajectory, PVT]:
        """Parse a trajectory from a gap.

        This function will:
          - Compute gap PVT points to bridge a previous and next collection window
          - Insert the previous collecion windows upper point into the trajectory
          - Insert the next collection windows first lower and midpoint
          - Calculate a 2 point average velocity for the first lower and midpoint
          - Produce an exit PVT of the next frames upper, midpoint-upper velocity, and
            time

        :param motor_info: Instance of _PmacMotorInfo
        :param gap: Index into slice where gap must occur
        :param motors: List of motors involved in trajectory
        :param slice: Information about a series of scan frames along a number of axes
        :param entry_pvt: PVT entering this trajectory
        :returns: Tuple of:
            Trajectory instance bridging previous and next collection windows with gap
            PVT at the start of the next collection window
        """
        slice_duration = error_if_none(slice.duration, "Slice must have a duration")
        half_durations = slice_duration / 2

        # Initialise exit PVT
        exit_pvt = PVT.default(motors)

        # Fill arrays for gap exit (start of next collection window)
        end_positions = {motor: np.empty(2, dtype=np.float64) for motor in motors}
        end_velocities = {motor: np.empty(2, dtype=np.float64) for motor in motors}
        end_durations = np.empty(2, dtype=np.float64)
        for motor in motors:
            end_positions[motor][0] = slice.lower[motor][gap]
            end_positions[motor][1] = slice.midpoints[motor][gap]

            mid_to_upper_velocity = (
                slice.upper[motor][gap] - slice.midpoints[motor][gap]
            ) / half_durations[gap]
            lower_to_mid_velocity = (
                slice.midpoints[motor][gap] - slice.lower[motor][gap]
            ) / half_durations[gap]

            end_velocities[motor][0] = lower_to_mid_velocity

            end_velocities[motor][1] = (
                lower_to_mid_velocity + mid_to_upper_velocity
            ) / 2
            end_durations[1] = half_durations[gap]

            exit_pvt.position[motor] = slice.upper[motor][gap]
            exit_pvt.velocity[motor] = mid_to_upper_velocity

        exit_pvt.time = half_durations[gap]

        # If we are ramping up, don't compute gap PVTs
        if ramp_up_time:
            end_durations[0] = ramp_up_time
            return Trajectory(
                positions=end_positions,
                velocities=end_velocities,
                durations=end_durations,
                user_programs=np.array([1, 1], dtype=int),
            ), exit_pvt

        entry_velocities = entry_pvt.velocity
        exit_velocities = {
            motor: (slice.midpoints[motor][gap] - slice.lower[motor][gap])
            / half_durations[gap]
            for motor in motors
        }
        distances = {
            motor: 0.0
            if np.isclose(
                (distance := slice.lower[motor][gap] - entry_pvt.position[motor]),
                0.0,
                atol=1e-12,
            )
            else distance
            for motor in motors
        }

        # Get velocity and time profiles across gap
        time_arrays, velocity_arrays = _get_velocity_profile(
            motors,
            motor_info,
            entry_velocities,
            exit_velocities,
            distances,
        )

        # Calculate gap PVTs
        gap_positions, gap_velocities, gap_durations = (
            _calculate_profile_from_velocities(
                motors,
                entry_pvt,
                time_arrays,
                velocity_arrays,
            )
        )

        # Initialise larger arrays for last collection windows
        # final upper point, the gap points, and the next collection windows
        # initial lower and midpoints
        # gap_duration includes duration for next collection windows initial
        # lower point, so array_size = len(gap_duration) + 2
        positions = {
            motor: np.empty(len(gap_durations) + 2, dtype=np.float64)
            for motor in motors
        }
        velocities = {
            motor: np.empty(len(gap_durations) + 2, dtype=np.float64)
            for motor in motors
        }
        durations = np.empty(len(gap_durations) + 2, dtype=np.float64)

        for motor in motors:
            # Insert last collection windows upper point
            positions[motor][0] = entry_pvt.position[motor]
            velocities[motor][0] = entry_pvt.velocity[motor]
            durations[0] = entry_pvt.time

            # Insert gap information
            positions[motor][1:-2] = gap_positions[motor]
            velocities[motor][1:-2] = gap_velocities[motor]
            durations[1:-1] = gap_durations

            # Append first 2 points of next collection window
            positions[motor][-2:] = end_positions[motor]
            velocities[motor][-2:] = end_velocities[motor]
            durations[-1] = end_durations[-1]

        trajectory = Trajectory(
            positions=positions,
            velocities=velocities,
            durations=durations,
            user_programs=np.array(
                [1] + [2] * (len(gap_durations) - 1) + [1, 1], dtype=int
            ),
        )

        return trajectory, exit_pvt


def _get_velocity_profile(
    motors: list[Motor],
    motor_info: _PmacMotorInfo,
    start_velocities: dict[Motor, np.float64],
    end_velocities: dict[Motor, np.float64],
    distances: dict[Motor, float],
) -> tuple[dict[Motor, npt.NDArray[np.float64]], dict[Motor, npt.NDArray[np.float64]]]:
    """Generate time and velocity profiles for motors across a gap.

    For each motor, a `VelocityProfile` is constructed.
    Profiles are iteratively recalculated to converge on a
    consistent minimum total gap time across all motors.

    This function will:
      - Initialise with a minimum turnaround time (`MIN_TURNAROUND`).
      - Build a velocity profile for each motor and determine the total
        move time required.
      - Update the minimum total gap time to the maximum of these totals.
      - Repeat until all profiles agree on the same minimum time or an
        iteration limit (i.e., 2) is reached.

    :param motors: Sequence of motors involved in trajectory
    :param motor_info: Instance of _PmacMotorInfo
    :param start_velocities: Motor velocities at start of gap
    :param end_velocities: Motor velocities at end of gap
    :param distances: Motor distances required to travel accross gap
    :raises ValueError: Cannot converge on common minimum time in 2 iterations
    :returns tuple: A tuple containing:
        dict: Motor's absolute timestamps of their velocity changes
        dict: Motor's velocity changes
    """
    profiles: dict[Motor, VelocityProfile] = {}
    time_arrays = {}
    velocity_arrays = {}

    min_time = MIN_TURNAROUND

    iterations = 2

    while iterations > 0:
        new_min_time = 0.0  # reset for this iteration

        for motor in motors:
            # Build profile for this motor
            p = VelocityProfile(
                start_velocities[motor],
                end_velocities[motor],
                distances[motor],
                min_time,
                motor_info.motor_acceleration_rate[motor],
                motor_info.motor_max_velocity[motor],
                0,
                MIN_INTERVAL,
            )
            p.get_profile()

            profiles[motor] = p
            new_min_time = max(new_min_time, p.t_total)

        # Check if all profiles have converged on min_time
        if np.isclose(new_min_time, min_time):
            for motor in motors:
                # MIN_INTERVAL should be less than our convergence tolerance
                # such that motors snap to the same point in the time grid
                profiles[motor].quantize()
                time_arrays[motor], velocity_arrays[motor] = profiles[
                    motor
                ].make_arrays()
            return time_arrays, velocity_arrays
        else:
            min_time = new_min_time
            iterations -= 1  # Get profiles with new minimum turnaround

    raise ValueError(
        "Failed to get a consistent time when calculating velocity profiles."
    )


def _calculate_profile_from_velocities(
    motors: list[Motor],
    entry_pvt: PVT,
    time_arrays: dict[Motor, npt.NDArray[np.float64]],
    velocity_arrays: dict[Motor, npt.NDArray[np.float64]],
) -> tuple[
    dict[Motor, npt.NDArray[np.float64]],
    dict[Motor, npt.NDArray[np.float64]],
    list[float],
]:
    """Convert per-axis time/velocity profiles into aligned time/position profiles.

    Given per-axis arrays of times and corresponding velocities,
    this builds a single unified timeline containing all unique velocity change points
    from every axis. It then steps through that timeline, and for each axis:

    * If the current time matches one of the axis's own velocity change points,
        use that known velocity.
    * Otherwise, linearly interpolate between the axis's previous and next
        known velocities, based on how far through that section we are.

    At each unified time step, the velocity is integrated over the step duration
    (using the trapezoidal rule) to update the axis's position. This produces
    per-axis position and velocity arrays that are aligned to the same global
    time grid.

    Example:
        combined_times = [0.1, 0.2, 0.3, 0.4]
        axis_times[motor] = [0.1, 0.4]
        velocity_array[motor] = [2, 5]

        At 0.1 → known vel = 2 (use directly)
        At 0.2 → 1/3 of the way to next vel → 3.0
        At 0.3 → 2/3 of the way to next vel → 4.0
        At 0.4 → known vel = 5 (use directly)

        These instantaneous velocities are integrated over each Δt to yield
        positions aligned with the global timeline.

    :param motors: Sequence of motors involved in trajectory
    :param slice: Information about a series of scan frames along a number of axes
    :param gap: Index into slice where gap has occured
    :param time_arrays: Motor's absolute timestamps of velocity changes
    :param velocity_arrays: Motor's velocity changes
    :returns GapSegment: Class representing a segment of a trajectory that is a gap
    """
    # Combine all per-axis time points into a single sorted array of times
    # This way we can evaluate each motor along the same timeline
    # We know all axes positions at t=0, so we drop this point
    combined_times = np.sort(np.unique(np.concatenate(list(time_arrays.values()))))[1:]

    # We convert a list of t into a list of Δt
    # We do this by substracting against previous cumulative time
    time_intervals = np.diff(np.concatenate(([0.0], combined_times))).tolist()

    # We also know all axes positions when t=t_final, so we drop this point
    # However, we need the interval for the next collection window, so we store it
    *time_intervals, final_interval = time_intervals
    combined_times = combined_times[:-1]
    num_intervals = len(time_intervals)

    # Prepare dicts for the resulting position and velocity profiles over the gap
    positions: dict[Motor, npt.NDArray[np.float64]] = {}
    velocities: dict[Motor, npt.NDArray[np.float64]] = {}

    # Loop over each motor and integrate its velocity profile over the unified times
    for motor in motors:
        axis_times = time_arrays[motor]
        axis_velocities = velocity_arrays[motor]
        axis_position = entry_pvt.position[
            motor
        ]  # start position at beginning of the gap
        prev_interval_vel = axis_velocities[
            0
        ]  # last velocity seen from the previous global time interval
        time_since_prev_axis_point = 0.0  # elapsed time since the last velocity point
        axis_idx = 1  # index into this axis's velocity/time arrays

        # Allocate output arrays for this motor with correct size
        positions[motor] = np.empty(num_intervals, dtype=np.float64)
        velocities[motor] = np.empty(num_intervals, dtype=np.float64)

        # Step through each interval in the Δt list
        for i, dt in enumerate(time_intervals):
            next_vel = axis_velocities[axis_idx]
            prev_vel = axis_velocities[axis_idx - 1]
            axis_dt = axis_times[axis_idx] - axis_times[axis_idx - 1]

            if np.isclose(combined_times[i], axis_times[axis_idx]):
                # If the current combined time exactly matches this motor's
                # next velocity change point, no interpolation is needed, so:
                this_vel = next_vel
                axis_idx += 1
                time_since_prev_axis_point = 0.0
            else:
                # Otherwise, linearly interpolate velocity between the previous
                # and next known velocity points for this motor
                time_since_prev_axis_point += dt
                # The fraction of the way we are from previous to next known velocities
                # for this motor
                frac = time_since_prev_axis_point / axis_dt
                # Interpolate for our velocity
                this_vel = prev_vel + frac * (next_vel - prev_vel)

            # Integrate velocity over this interval to update position.
            # Using the trapezoidal rule:
            delta_pos = 0.5 * (prev_interval_vel + this_vel) * dt
            axis_position += delta_pos
            prev_interval_vel = this_vel  # update for next loop

            # Store the computed position and velocity for this interval
            positions[motor][i] = axis_position
            velocities[motor][i] = this_vel

    return (
        positions,
        velocities,
        time_intervals + [final_interval],
    )



import numpy as np
import pytest
from scanspec.core import Path
from scanspec.specs import Fly, Line, Spiral

from ophyd_async.epics.motor import Motor
from ophyd_async.epics.pmac import (
    PmacIO,
)
from ophyd_async.epics.pmac._pmac_trajectory_generation import (
    MIN_INTERVAL,  # noqa: PLC2701
    PVT,  # noqa: PLC2701
    Trajectory,  # noqa: PLC2701
)
from ophyd_async.epics.pmac._utils import (
    _PmacMotorInfo,  # noqa: PLC2701
)
from ophyd_async.testing import set_mock_value


@pytest.fixture
async def motor_info(sim_motors: tuple[PmacIO, Motor, Motor]):
    _, sim_x_motor, sim_y_motor = sim_motors
    yield _PmacMotorInfo(
        "CS1",
        1,
        {sim_x_motor: 6, sim_y_motor: 7},
        {sim_x_motor: 10, sim_y_motor: 10},
        {sim_x_motor: 5, sim_y_motor: 5},
    )


async def test_line_trajectory_from_slice(
    sim_motors: tuple[PmacIO, Motor, Motor], motor_info: _PmacMotorInfo
):
    _, sim_x_motor, _ = sim_motors
    spec = Fly(2.0 @ Line(sim_x_motor, 1, 5, 9))
    slice = Path(spec.calculate()).consume()
    trajectory, exit_pvt = Trajectory.from_slice(slice, motor_info, ramp_up_time=2)
    trajectory = trajectory.with_ramp_down(exit_pvt, {sim_x_motor: np.float64(6)}, 2, 0)

    assert trajectory.positions[sim_x_motor] == pytest.approx(
        [
            0.75,
            1.0,
            1.25,
            1.5,
            1.75,
            2.0,
            2.25,
            2.5,
            2.75,
            3.0,
            3.25,
            3.5,
            3.75,
            4.0,
            4.25,
            4.5,
            4.75,
            5.0,
            5.25,
            6.0,
        ]
    )

    assert trajectory.velocities[sim_x_motor] == pytest.approx(
        [
            0.25,
            0.25,
            0.25,
            0.25,
            0.25,
            0.25,
            0.25,
            0.25,
            0.25,
            0.25,
            0.25,
            0.25,
            0.25,
            0.25,
            0.25,
            0.25,
            0.25,
            0.25,
            0.25,
            0.0,
        ]
    )

    assert (
        trajectory.user_programs
        == [
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            8,
        ]
    ).all()

    assert trajectory.durations == pytest.approx(
        [
            2.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            2.0,
        ],
    )


async def test_spiral_trajectory_from_slice(
    sim_motors: tuple[PmacIO, Motor, Motor], motor_info: _PmacMotorInfo
):
    _, sim_x_motor, sim_y_motor = sim_motors
    spec = Spiral(sim_x_motor, sim_y_motor, 0, 0, 5, 5, 3)
    slice = Path(Fly(2.0 @ spec).calculate()).consume()

    trajectory, exit_pvt = Trajectory.from_slice(slice, motor_info, ramp_up_time=2.0)
    trajectory = trajectory.with_ramp_down(
        exit_pvt, {sim_x_motor: np.float64(0), sim_y_motor: np.float64(0)}, 2.0, 0
    )

    assert trajectory.positions == {
        sim_x_motor: pytest.approx(
            [
                0.0,
                0.60538,
                -0.56648101,
                -1.64763742,
                -1.94954837,
                -1.43181015,
                -0.35683972,
                0.0,
            ]
        ),
        sim_y_motor: pytest.approx(
            [
                0.0,
                -0.82169442,
                -1.32756642,
                -0.64053956,
                0.60491969,
                1.77714744,
                2.47440203,
                0.0,
            ]
        ),
    }

    assert trajectory.velocities == {
        sim_x_motor: pytest.approx(
            [
                0.60538,
                -0.28324051,
                -1.12650871,
                -0.69153368,
                0.10791364,
                0.79635432,
                1.07497043,
                0.0,
            ],
        ),
        sim_y_motor: pytest.approx(
            [
                -0.82169442,
                -0.66378321,
                0.09057743,
                0.96624305,
                1.2088435,
                0.93474117,
                0.69725459,
                0.0,
            ],
        ),
    }

    assert trajectory.durations == pytest.approx(
        [2.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 2.0]
    )

    assert (trajectory.user_programs == [1, 1, 1, 1, 1, 1, 1, 8]).all()


async def test_snaked_trajectory_with_gaps(
    sim_motors: tuple[PmacIO, Motor, Motor], motor_info: _PmacMotorInfo
):
    _, sim_x_motor, sim_y_motor = sim_motors
    spec = Fly(1.0 @ (Line(sim_y_motor, 10, 12, 3) * ~Line(sim_x_motor, 1, 5, 5)))
    slice = Path(spec.calculate()).consume()

    trajectory, exit_pvt = Trajectory.from_slice(slice, motor_info, ramp_up_time=1.0)
    trajectory = trajectory.with_ramp_down(
        exit_pvt, {sim_x_motor: np.float64(6.0), sim_y_motor: np.float64(12)}, 1.0, 0
    )

    assert trajectory.positions[sim_x_motor] == pytest.approx(
        [
            0.5,
            1.0,
            1.5,
            2.0,
            2.5,
            3.0,
            3.5,
            4.0,
            4.5,
            5.0,
            5.5,
            5.55,
            5.55,
            5.55,
            5.5,
            5.0,
            4.5,
            4.0,
            3.5,
            3.0,
            2.5,
            2.0,
            1.5,
            1.0,
            0.5,
            0.45,
            0.45,
            0.45,
            0.5,
            1.0,
            1.5,
            2.0,
            2.5,
            3.0,
            3.5,
            4.0,
            4.5,
            5.0,
            5.5,
            6.0,
        ]
    )

    assert trajectory.velocities[sim_x_motor] == pytest.approx(
        [
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            -0.0,
            0.0,
            -0.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -0.0,
            0.0,
            -0.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            0.0,
        ]
    )

    assert trajectory.positions[sim_y_motor] == pytest.approx(
        [
            10.0,
            10.0,
            10.0,
            10.0,
            10.0,
            10.0,
            10.0,
            10.0,
            10.0,
            10.0,
            10.0,
            10.05,
            10.5,
            10.95,
            11.0,
            11.0,
            11.0,
            11.0,
            11.0,
            11.0,
            11.0,
            11.0,
            11.0,
            11.0,
            11.0,
            11.05,
            11.5,
            11.95,
            12.0,
            12.0,
            12.0,
            12.0,
            12.0,
            12.0,
            12.0,
            12.0,
            12.0,
            12.0,
            12.0,
            12.0,
        ],
        1e-3,
    )

    assert trajectory.velocities[sim_y_motor] == pytest.approx(
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.976,
            3.125,
            0.976,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.977,
            3.125,
            0.977,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
        1e-3,
    )

    assert (
        trajectory.user_programs
        == [
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            2,
            2,
            2,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            2,
            2,
            2,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            8,
        ]
    ).all()

    assert trajectory.durations == pytest.approx(
        [
            1.0,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.1,
            0.22,
            0.22,
            0.1,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.1,
            0.22,
            0.22,
            0.1,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            1.0,
        ],
        1e-5,
    )


async def test_grid_trajectory_with_gaps(
    sim_motors: tuple[PmacIO, Motor, Motor], motor_info: _PmacMotorInfo
):
    _, sim_x_motor, sim_y_motor = sim_motors
    spec = Fly(1.0 @ (Line(sim_y_motor, 10, 12, 3) * Line(sim_x_motor, 1, 5, 5)))
    slice = Path(spec.calculate()).consume()

    trajectory, exit_pvt = Trajectory.from_slice(slice, motor_info, ramp_up_time=1.0)
    trajectory = trajectory.with_ramp_down(
        exit_pvt,
        {sim_x_motor: np.float64(6.0), sim_y_motor: np.float64(12.0)},
        1.0,
        0.0,
    )

    assert trajectory.positions[sim_x_motor] == pytest.approx(
        [
            0.5,
            1.0,
            1.5,
            2.0,
            2.5,
            3.0,
            3.5,
            4.0,
            4.5,
            5.0,
            5.5,
            5.5421,
            4.3053,
            1.6947,
            0.4572,
            0.5,
            1.0,
            1.5,
            2.0,
            2.5,
            3.0,
            3.5,
            4.0,
            4.5,
            5.0,
            5.5,
            5.5421,
            4.3053,
            1.6947,
            0.4572,
            0.5,
            1.0,
            1.5,
            2.0,
            2.5,
            3.0,
            3.5,
            4.0,
            4.5,
            5.0,
            5.5,
            6,
        ],
        1e-3,
    )

    assert trajectory.velocities[sim_x_motor] == pytest.approx(
        [
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            0.3818,
            -4.9822,
            -4.9822,
            0.3818,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            0.3818,
            -4.9822,
            -4.9822,
            0.3818,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            0.0,
        ],
        1e-3,
    )

    assert trajectory.positions[sim_y_motor] == pytest.approx(
        [
            10.0,
            10.0,
            10.0,
            10.0,
            10.0,
            10.0,
            10.0,
            10.0,
            10.0,
            10.0,
            10.0,
            10.01865,
            10.34235,
            10.65764,
            10.98134,
            11.0,
            11.0,
            11.0,
            11.0,
            11.0,
            11.0,
            11.0,
            11.0,
            11.0,
            11.0,
            11.0,
            11.01865,
            11.34235,
            11.65764,
            11.98134,
            12.0,
            12.0,
            12.0,
            12.0,
            12.0,
            12.0,
            12.0,
            12.0,
            12.0,
            12.0,
            12.0,
            12.0,
        ]
    )

    assert trajectory.velocities[sim_y_motor] == pytest.approx(
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.601684717208,
            0.601684717208,
            0.601684717208,
            0.601684717208,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.601684717208,
            0.601684717208,
            0.601684717208,
            0.601684717208,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ]
    )

    assert (
        trajectory.user_programs
        == [
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            2,
            2,
            2,
            2,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            2,
            2,
            2,
            2,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            8,
        ]
    ).all()

    assert trajectory.durations == pytest.approx(
        [
            1.0,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.062,
            0.538,
            0.524,
            0.538,
            0.062,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.062,
            0.538,
            0.524,
            0.538,
            0.062,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            1.0,
        ],
        1e-5,
    )


async def test_from_gap(
    sim_motors: tuple[PmacIO, Motor, Motor], motor_info: _PmacMotorInfo
):
    _, sim_x_motor, sim_y_motor = sim_motors
    spec = Fly(1.0 @ (Line(sim_y_motor, 10, 12, 3) * ~Line(sim_x_motor, 1, 5, 5)))
    slice = Path(spec.calculate()).consume()

    # This is info about the frame just before a gap is created.
    entry_pvt = PVT(
        position={sim_x_motor: np.float64(5.5), sim_y_motor: np.float64(10.0)},
        velocity={sim_x_motor: np.float64(1.0), sim_y_motor: np.float64(0.0)},
        time=np.float64(1),
    )

    trajectory, _ = Trajectory.from_gap(
        motor_info, 5, [sim_x_motor, sim_y_motor], slice, entry_pvt
    )

    assert trajectory.positions[sim_x_motor] == pytest.approx(
        [
            5.5,
            5.55,
            5.55,
            5.55,
            5.5,
            5.0,
        ],
    )

    assert trajectory.positions[sim_y_motor] == pytest.approx(
        [
            10.0,
            10.05,
            10.5,
            10.95,
            11.0,
            11.0,
        ],
        1e-3,
    )

    assert trajectory.velocities[sim_x_motor] == pytest.approx(
        [
            1.0,
            -0.0,
            0.0,
            -0.0,
            -1.0,
            -1.0,
        ],
    )

    assert trajectory.velocities[sim_y_motor] == pytest.approx(
        [
            0.0,
            0.9765,
            3.125,
            0.9765,
            0.0,
            0.0,
        ],
        1e-3,
    )

    assert trajectory.durations == pytest.approx(
        [
            1.0,
            0.1,
            0.22,
            0.22,
            0.1,
            0.5,
        ]
    )

    assert (trajectory.user_programs == [1, 2, 2, 2, 1, 1]).all()


async def test_from_collection_window(sim_motors: tuple[PmacIO, Motor, Motor]):
    _, _, sim_y_motor = sim_motors
    spec = Fly(1.0 @ (Line(sim_y_motor, 1, 5, 5)))
    path = Path(spec.calculate())
    slice = path.consume()

    entry_pvt = PVT(
        position={sim_y_motor: np.float64(1.0)},
        velocity={sim_y_motor: np.float64(0.5)},
        time=np.float64(0.5),
    )

    trajectory, _ = Trajectory.from_collection_window(
        1, 5, [sim_y_motor], slice, entry_pvt
    )

    assert trajectory.positions[sim_y_motor] == pytest.approx(
        [1.5, 2, 2.5, 3.0, 3.5, 4.0, 4.5, 5.0],
    )

    assert trajectory.velocities[sim_y_motor] == pytest.approx(
        [0.75, 1, 1, 1, 1, 1, 1, 1],
    )

    assert trajectory.durations == pytest.approx(
        [0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5],
    )

    assert (trajectory.user_programs == [1, 1, 1, 1, 1, 1, 1, 1]).all()


async def test_appending_trajectory(
    sim_motors: tuple[PmacIO, Motor, Motor], motor_info: _PmacMotorInfo
):
    _, sim_x_motor, sim_y_motor = sim_motors
    spec = Fly(1.0 @ (Line(sim_y_motor, 10, 11, 2) * ~Line(sim_x_motor, 1, 3, 3)))
    path = Path(spec.calculate())
    slice = path.consume(3)

    first_trajectory, first_exit_pvt = Trajectory.from_slice(
        slice, motor_info, ramp_up_time=1
    )

    slice = path.consume()
    second_trajectory, second_exit_pvt = Trajectory.from_slice(
        slice, motor_info, first_exit_pvt
    )

    overall_trajectory = Trajectory.from_trajectories(
        [first_trajectory, second_trajectory], [sim_x_motor, sim_y_motor]
    )

    overall_trajectory = overall_trajectory.with_ramp_down(
        second_exit_pvt,
        {sim_x_motor: np.float64(0), sim_y_motor: np.float64(11.0)},
        1.0,
        0.0,
    )

    assert overall_trajectory.positions[sim_x_motor] == pytest.approx(
        [
            0.5,
            1.0,
            1.5,
            2.0,
            2.5,
            3.0,
            3.5,
            3.55,
            3.55,
            3.55,
            3.5,
            3.0,
            2.5,
            2.0,
            1.5,
            1.0,
            0.5,
            0.0,
        ]
    )

    assert overall_trajectory.positions[sim_y_motor] == pytest.approx(
        [
            10,
            10,
            10,
            10,
            10,
            10,
            10,
            10.05,
            10.50,
            10.95,
            11.0,
            11.0,
            11.0,
            11.0,
            11.0,
            11.0,
            11.0,
            11.0,
        ],
        1e-3,
    )

    assert overall_trajectory.velocities[sim_x_motor] == pytest.approx(
        [
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            1.0,
            -0.0,
            0.0,
            -0.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            -1.0,
            0.0,
        ]
    )

    assert overall_trajectory.velocities[sim_y_motor] == pytest.approx(
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.9765,
            3.125,
            0.9765,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
        1e-3,
    )

    assert overall_trajectory.durations == pytest.approx(
        [
            1.0,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.1,
            0.22,
            0.22,
            0.1,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            0.5,
            1.0,
        ],
        1e-5,
    )

    assert (
        overall_trajectory.user_programs
        == [
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            2,
            2,
            2,
            1,
            1,
            1,
            1,
            1,
            1,
            1,
            8,
        ]
    ).all()


async def test_hardware_triggered_step_scan(
    sim_motors: tuple[PmacIO, Motor, Motor], motor_info: _PmacMotorInfo
):
    _, sim_x_motor, _ = sim_motors
    spec = 1.0 @ Line(sim_x_motor, 1, 3, 3)
    path = Path(spec.calculate())
    slice = path.consume()

    trajectory, exit_pvt = Trajectory.from_slice(slice, motor_info, ramp_up_time=2)
    trajectory = trajectory.with_ramp_down(exit_pvt, {sim_x_motor: np.float64(6)}, 2, 0)

    assert trajectory.positions[sim_x_motor] == pytest.approx(
        [
            1.0,
            1.0,
            1.0,
            1.5,
            2.0,
            2.0,
            2.0,
            2.5,
            3.0,
            3.0,
            3.0,
            6.0,
        ]
    )

    assert trajectory.velocities[sim_x_motor] == pytest.approx(
        [
            0.0,
            0.0,
            0.0,
            3.125,
            0.0,
            0.0,
            0.0,
            3.125,
            0.0,
            0.0,
            0.0,
            0.0,
        ],
        1e-5,
    )

    assert trajectory.durations == pytest.approx(
        [
            2.0,
            0.5,
            0.5,
            0.32,
            0.32,
            0.5,
            0.5,
            0.32,
            0.32,
            0.5,
            0.5,
            2.0,
        ],
        1e-5,
    )

    assert (
        trajectory.user_programs
        == [
            1,
            1,
            1,
            2,
            1,
            1,
            1,
            2,
            1,
            1,
            1,
            8,
        ]
    ).all()


async def test_small_turnaround_durations_are_quantized(
    sim_motors: tuple[PmacIO, Motor, Motor],
):
    pmac, sim_x_motor, sim_y_motor = sim_motors

    # These parameters result in gap point durations < MIN_INTERVAL
    set_mock_value(sim_y_motor.max_velocity, 100)
    set_mock_value(sim_y_motor.acceleration_time, 0.25)
    motor_info = await _PmacMotorInfo.from_motors(pmac, [sim_x_motor, sim_y_motor])
    spec = Fly(float(1) @ (Line(sim_y_motor, 0, 1, 10) * ~Line(sim_x_motor, 0, 1, 10)))
    slice = Path(spec.calculate()).consume()

    # Parse the first turnaround profile
    trajectory, _ = Trajectory.from_gap(
        motor_info=motor_info,
        gap=10,
        motors=[sim_x_motor, sim_y_motor],
        slice=slice,
        entry_pvt=PVT(
            position={sim_x_motor: np.float64(1.055), sim_y_motor: np.float64(0)},
            velocity={sim_x_motor: np.float64(0.1111), sim_y_motor: np.float64(0)},
            time=np.float64(0.5),
        ),
    )

    # Check that gap durations are snapped up to nearest interval
    assert min(trajectory.durations) == pytest.approx(MIN_INTERVAL)
